#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from ...domains.materials import MacroMolecule
from ..utils import get_descriptor


def length(protein: MacroMolecule) -> int:
    """Retrieves the number of residues of a protein."""
    desc = get_descriptor(protein)
    desc.length()
    return int(desc.descriptor)


def molecular_weight(protein: MacroMolecule, amide: bool = False) -> float:
    """Computes the molecular weight of a protein."""
    desc = get_descriptor(protein)
    desc.calculate_MW(amide=amide)
    return float(desc.descriptor)


def boman_index(protein: MacroMolecule) -> float:
    """Computes the Boman index of a protein (sum of solubility values of all residues).

    Boman, H. G. (2003).
    Antibacterial peptides: basic facts and emerging concepts.
    Journal of internal medicine, 254(3), 197-215.
    """
    desc = get_descriptor(protein)
    desc.boman_index()
    return float(desc.descriptor)


def aliphatic_index(protein: MacroMolecule) -> float:
    """Computes the aliphatic index of a protein. Measure of thermal stability.

    Ikai, A. (1980).
    Thermostability and aliphatic index of globular proteins.
    The Journal of Biochemistry, 88(6), 1895-1898.
    """
    desc = get_descriptor(protein)
    desc.aliphatic_index()
    return float(desc.descriptor)


def hydrophobic_ratio(protein: MacroMolecule) -> float:
    """Computes the hydrophobicity of a protein, relative freq. of **A,C,F,I,L,M & V**."""
    desc = get_descriptor(protein)
    desc.hydrophobic_ratio()
    return float(desc.descriptor)


def charge(protein: MacroMolecule, ph: float = 7.0, amide: bool = False) -> float:
    """Computes the charge of a protein.

    Bjellqvist, B., Hughes, G. J., Pasquali, C., Paquet, N., Ravier, F., Sanchez, J. C., ... & Hochstrasser, D. (1993).
    The focusing positions of polypeptides in immobilized pH gradients can be predicted from their amino acid sequences.
    Electrophoresis, 14(1), 1023-1031.
    """
    desc = get_descriptor(protein)
    desc.calculate_charge(ph=ph, amide=amide)
    return float(desc.descriptor)


def charge_density(
    protein: MacroMolecule, ph: float = 7.0, amide: bool = False
) -> float:
    """Computes the charge density of a protein.

    Bjellqvist, B., Hughes, G. J., Pasquali, C., Paquet, N., Ravier, F., Sanchez, J. C., ... & Hochstrasser, D. (1993).
    The focusing positions of polypeptides in immobilized pH gradients can be predicted from their amino acid sequences.
    Electrophoresis, 14(1), 1023-1031.
    """
    desc = get_descriptor(protein)
    desc.charge_density(ph=ph, amide=amide)
    return float(desc.descriptor)


def isoelectric_point(protein: MacroMolecule, amide: bool = False) -> float:
    """Computes the isoelectric point of every residue and aggregates."""
    desc = get_descriptor(protein)
    desc.isoelectric_point(amide=amide)
    return float(desc.descriptor)


def aromaticity(protein: MacroMolecule) -> float:
    """Computes aromaticity of the protein (relative frequency of Phe+Trp+Tyr).

    Lobry, J. R., & Gautier, C. (1994).
    Hydrophobicity, expressivity and aromaticity are the major trends of amino-acid usage
        in 999 Escherichia coli chromosome-encoded genes.
    Nucleic acids research, 22(15), 3174-3180.
    """
    desc = get_descriptor(protein)
    desc.aromaticity()
    return float(desc.descriptor)


def instability(protein: MacroMolecule) -> float:
    """Calculates the protein instability.

    Guruprasad, K., Reddy, B. B., & Pandit, M. W. (1990).
    Correlation between stability of a protein and its dipeptide composition: a novel
        approach for predicting in vivo stability of a protein from its primary sequence.
    Protein Engineering, Design and Selection, 4(2), 155-161.
    """
    desc = get_descriptor(protein)
    desc.instability_index()
    return float(desc.descriptor)



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Types, classes, validation, etc. for the material domain."""

from typing import List, NewType, Tuple, Union

import numpy as np
import pandas as pd
from rdkit import Chem
from rdkit.Chem import Mol

# TODO: setting to str directly requires no wrapping, so wrong strings could be passed
Protein = str  # NewType('Protein', str)
SMILES = str  # NewType('SMILES', str)
SmallMolecule = Union[SMILES, Mol]
MacroMolecule = Union[Protein, Mol]
Omics = Union[np.ndarray, pd.Series]
PAG = SMILES
Molecule = Union[SmallMolecule, MacroMolecule]
Sequence = str


def validate_molecules(
    smiles_list: List[SMILES],
) -> Tuple[List[Chem.rdchem.Mol], List[int]]:
    """Validate molecules.

    Args:
        smiles_list: list of SMILES representing molecules.

    Returns:
        a tuple containing RDKit molecules and valid indexes.
    """
    # generate molecules from SMILES
    molecules = [
        Chem.MolFromSmiles(a_smiles, sanitize=True) for a_smiles in smiles_list
    ]
    # valid ids
    valid_ids = [
        index
        for index, molecule in enumerate(molecules)
        if molecule is not None and molecule != ""
    ]
    return molecules, valid_ids


Bounds = Tuple[int, int]  # NewType('Bounds', Tuple[int, int])

PhotoacidityCondition = NewType("PhotoacidityCondition", Bounds)
# photoacidity_condition = PhotoacidityCondition(
#     (0, 1)
# )  # PhotoacidityCondition(Bounds((0, 1))) if Bound was a new type

ConditionPAG = Union[PhotoacidityCondition]



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from typing import Any, Dict, List

from .core import PropertyPredictor
from .molecules import MOLECULE_PROPERTY_PREDICTOR_FACTORY
from .proteins import PROTEIN_PROPERTY_PREDICTOR_FACTORY
from .scorer import (
    MoleculePropertyPredictorScorer,
    PropertyPredictorScorer,
    ProteinPropertyPredictorScorer,
)

PROPERTY_PREDICTOR_FACTORY: Dict[str, Any] = {
    **MOLECULE_PROPERTY_PREDICTOR_FACTORY,
    **PROTEIN_PROPERTY_PREDICTOR_FACTORY,
}

AVAILABLE_PROPERTY_PREDICTORS = sorted(PROPERTY_PREDICTOR_FACTORY.keys())

SCORING_FACTORY_WITH_PROPERTY_PREDICTORS = {
    "property_predictor_scorer": PropertyPredictorScorer,
    "molecule_property_predictor_scorer": MoleculePropertyPredictorScorer,
    "protein_property_predictor_scorer": ProteinPropertyPredictorScorer,
}

AVAILABLE_SCORING_WITH_PROPERTY_PREDICTORS = sorted(
    SCORING_FACTORY_WITH_PROPERTY_PREDICTORS.keys()
)


class PropertyPredictorRegistry:
    """A registry for property predictors."""

    @staticmethod
    def get_property_predictor_parameters_schema(name: str) -> Dict[str, Any]:
        try:
            _, parameters_class = PROPERTY_PREDICTOR_FACTORY[name]
            return parameters_class.schema_json()
        except KeyError:
            raise ValueError(
                f"Property predictor name={name} not supported. Pick one from {AVAILABLE_PROPERTY_PREDICTORS}"
            )

    @staticmethod
    def get_property_predictor(
        name: str, parameters: Dict[str, Any] = {}
    ) -> PropertyPredictor:
        try:
            property_class, parameters_class = PROPERTY_PREDICTOR_FACTORY[name]
            return property_class(parameters_class(**parameters))
        except KeyError:
            raise ValueError(
                f"Property predictor name={name} not supported. Pick one from {AVAILABLE_PROPERTY_PREDICTORS}"
            )

    @staticmethod
    def get_property_predictor_scorer(
        property_name: str,
        scorer_name: str,
        target: float,
        parameters: Dict[str, Any] = {},
    ) -> PropertyPredictorScorer:
        """Get a property predictor scorer.

        Args:
            property_name: name of the property to score.
            scorer_name: name of the scorer to use.
            target: target score that will be used to get the distance to the score of a molecule or protein (not be confused with parameters["target"]).
            parameters: parameters for the scoring function.

        Returns:
            A property predictor scorer.
        """
        scoring_function = PropertyPredictorRegistry.get_property_predictor(
            name=property_name, parameters=parameters
        )

        if scorer_name not in SCORING_FACTORY_WITH_PROPERTY_PREDICTORS:
            raise ValueError(
                f"Scorer name={scorer_name} not supported. Pick one from {AVAILABLE_SCORING_WITH_PROPERTY_PREDICTORS}"
            )
        property_predictor_scorer = SCORING_FACTORY_WITH_PROPERTY_PREDICTORS[
            scorer_name
        ]
        return property_predictor_scorer(property_name, scoring_function, target)

    @staticmethod
    def list_available() -> List[str]:
        return AVAILABLE_PROPERTY_PREDICTORS

    @staticmethod
    def list_available_scorers() -> List[str]:
        return AVAILABLE_SCORING_WITH_PROPERTY_PREDICTORS



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import argparse
import configparser
from typing import Any, Dict, Optional

import sentencepiece as _sentencepiece
import numpy as np
from pytorch_lightning import Trainer

from ..ml.models import MODEL_FACTORY
from .utils import convert_string_to_class

# sentencepiece has to be loaded before lightning to avoid segfaults
_sentencepiece


def parse_arguments_from_config(conf_file: Optional[str] = None) -> argparse.Namespace:
    """Parse arguments from configuration file.

    Args:
        conf_file: configuration file. Defaults to None, a.k.a. us a default configuration
            in ./config/config.ini.

    Returns:
        the parsed arguments.
    """
    parser = argparse.ArgumentParser()

    # open config.ini file, either from parser or default file
    parser.add_argument(
        "--conf_file",
        type=str,
        help=("config file for the defaults value"),
        default="./config/config.ini",
    )

    # Read config file
    args, remaining_argv = parser.parse_known_args()
    config = configparser.ConfigParser()

    if conf_file:
        config.read(conf_file)
    else:
        config.read(args.conf_file)

    # classes that are not model name
    general_config_classes = ["general", "trainer", "default"]

    # adding a list of all model name into the args
    result: Dict[str, Any] = dict()
    result["model_list"] = [
        i for i in list(config.keys()) if i.lower() not in general_config_classes
    ]
    for key in [*config.keys()]:
        # go trough all models parameter, replace the parsed ones from the the config files ones
        if key.lower() not in general_config_classes:
            model_type = config[key]["type"]
            params_from_configfile = dict(config[key])
            model = MODEL_FACTORY[model_type.lower()]
            parser = model.add_model_specific_args(parser, key)
            args, _ = parser.parse_known_args()
            args_dictionary = vars(args)
            params_from_configfile["name"] = key

            for i in params_from_configfile:
                params_from_configfile[i] = convert_string_to_class(
                    params_from_configfile[i]
                )

            params_from_configfile.update(
                {
                    k[: -len(key) - 1]: v
                    for k, v in args_dictionary.items()
                    if v is not None and k.endswith("_" + key)
                }
            )

            result[key] = params_from_configfile

        elif key.lower() == "trainer" or key.lower() == "general":
            params_from_configfile = dict(config[key])
            for i in params_from_configfile:
                params_from_configfile[i] = convert_string_to_class(
                    params_from_configfile[i]
                )
            result.update(params_from_configfile)

    # parser Pytorch Trainer arguments
    parser = Trainer.add_argparse_args(parser)

    # parse the arguments
    parser.add_argument("--algorithm", type=str, default="trajectory_balance")
    parser.add_argument("--context", type=str, default=None)
    parser.add_argument("--dataset", type=str, default="qm9")
    parser.add_argument("--dataset_path", type=str, default="./data/qm9.h5")
    parser.add_argument("--environment", type=str, default=None)
    parser.add_argument("--model", type=str, default="graph_transformer_gfn")
    parser.add_argument("--sampling_model", type=str, default="graph_transformer_gfn")
    parser.add_argument("--task", type=str, default="qm9")

    # adding basename as the name of the run
    parser.add_argument("--epoch", type=int)
    parser.add_argument("--basename", type=str)
    parser.add_argument("--batch_size", type=int, default=64)
    parser.add_argument("--num_workers", type=int, default=0)

    parser.add_argument("--validation_split", type=float)
    parser.add_argument("--validation_indices_file", type=str)
    parser.add_argument("--stratified_batch_file", type=str)
    parser.add_argument("--stratified_value_name", type=str)
    parser.add_argument("--checkpoint_every_n_val_epochs", type=str)
    parser.add_argument("--trainer_log_every_n_steps", type=str)
    parser.add_argument("--trainer_flush_logs_every_n_steps", type=str)
    parser.add_argument("--test_output_path", type=str, default="./test")

    parser.add_argument("--log_dir", type=str, default="./log/")
    parser.add_argument("--num_training_steps", type=int, default=1000)
    parser.add_argument("--validate_every", type=int, default=1000)

    parser.add_argument("--seed", type=int, default=142857)
    parser.add_argument("--device", type=str, default="cpu")

    # arguments specific for gflownet
    parser.add_argument("--bootstrap_own_reward", type=bool, default=False)
    parser.add_argument("--learning_rate", type=float, default=1e-4)
    parser.add_argument("--global_batch_size", type=int, default=16)
    parser.add_argument("--num_emb", type=int, default=128)
    parser.add_argument("--num_layers", type=int, default=4)
    parser.add_argument("--tb_epsilon", type=float, default=1e-10)
    parser.add_argument("--illegal_action_logreward", type=float, default=-50)
    parser.add_argument("--reward_loss_multiplier", type=float, default=1)
    parser.add_argument("--temperature_sample_dist", type=str, default="uniform")
    parser.add_argument("--temperature_dist_params", type=str, default="(.5, 32)")
    parser.add_argument("--weight_decay", type=float, default=1e-8)
    parser.add_argument("--num_data_loader_workers", type=float, default=8)
    parser.add_argument("--momentum", type=float, default=0.9)
    parser.add_argument("--adam_eps", type=float, default=1e-8)
    parser.add_argument("--lr_decay", type=int, default=20000)
    parser.add_argument("--z_lr_decay", type=int, default=20000)
    parser.add_argument("--clip_grad_type", type=str, default="norm")
    parser.add_argument("--clip_grad_param", type=int, default=10)
    parser.add_argument("--random_action_prob", type=float, default=0.001)
    parser.add_argument("--sampling_tau", type=float, default=0.0)
    parser.add_argument("--max_nodes", type=int, default=9)
    parser.add_argument("--num_offline", type=int, default=10)
    parser.add_argument("--sampling_iterator", type=bool, default=True)
    parser.add_argument("--ratio", type=float, default=0.9)
    parser.add_argument("--distributed_training_strategy", type=str, default="ddp")
    parser.add_argument("--development_mode", type=bool, default=False)

    args_dictionary = vars(parser.parse_args(remaining_argv))
    result.update({k: v for k, v in args_dictionary.items() if v is not None})
    # add random number generator from seed
    result["rng"] = np.random.default_rng(result["seed"])
    result_namespace = argparse.Namespace(**result)

    return result_namespace



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from .unpatch import (  # isort:skip
    fix_datasets,
    sane_datasets,
    fix_schedulers,
    sane_schedulers,
)

from torchdrug.datasets import (
    BACE,
    BBBP,
    CEP,
    HIV,
    MOSES,
    MUV,
    OPV,
    PCQM4M,
    QM8,
    QM9,
    SIDER,
    ChEMBLFiltered,
    ClinTox,
    Delaney,
    FreeSolv,
    Lipophilicity,
    Malaria,
    PubChem110m,
    Tox21,
    ToxCast,
    ZINC2m,
    ZINC250k,
)

from .dataset import TorchDrugDataset

# isort: off
from torch import nn

"""
Necessary because torchdrug silently overwrites the default nn.Module. This is quite
invasive and causes significant side-effects in the rest of the code.
See: https://github.com/DeepGraphLearning/torchdrug/issues/77
"""
nn.Module = nn._Module  # type: ignore
fix_datasets(sane_datasets)
fix_schedulers(sane_schedulers)


DATASET_FACTORY = {
    "bace": BACE,
    "bbbp": BBBP,
    "custom": TorchDrugDataset,
    "cep": CEP,
    "chembl": ChEMBLFiltered,
    "clintox": ClinTox,
    "delaney": Delaney,
    "freesolv": FreeSolv,
    "hiv": HIV,
    "lipophilicity": Lipophilicity,
    "malaria": Malaria,
    "moses": MOSES,
    "muv": MUV,
    "opv": OPV,
    "pcqm4m": PCQM4M,
    "pubchem": PubChem110m,
    "qm8": QM8,
    "qm9": QM9,
    "sider": SIDER,
    "tox21": Tox21,
    "toxcast": ToxCast,
    "zinc250k": ZINC250k,
    "zinc2m": ZINC2m,
}



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Advanced manufacturing algorithms."""

import logging
from dataclasses import field
from typing import Any, Callable, ClassVar, Dict, Iterable, Optional, TypeVar

from ....domains.materials import SmallMolecule
from ....training_pipelines.core import TrainingPipelineArguments
from ....training_pipelines.pytorch_lightning.granular.core import (
    GranularSavingArguments,
)
from ...core import AlgorithmConfiguration, GeneratorAlgorithm
from ...registry import ApplicationsRegistry
from .implementation.core import Generator
from .implementation.nccr import CatalystGenerator as NCCRCatalystGenerator

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

T = TypeVar("T", bound=Any)
S = TypeVar("S", bound=Any)
Targeted = Callable[[T], Iterable[Any]]


class AdvancedManufacturing(GeneratorAlgorithm[S, T]):
    """Advance manufacturing generator algorithm."""

    def __init__(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ):
        """Instantiate AdvancedManufacturing ready to generate items.

        Args:
            configuration: domain and application
                specification defining parameters, types and validations.
            target: a target for which to generate items.

        Example:
            An example for generating small molecules (SMILES) with a target binding energy::

                config = CatalystGenerator()
                algorithm = AdvancedManufacturing(configuration=config, target=10.0)
                items = list(algorithm.sample(10))
                print(items)
        """

        configuration = self.validate_configuration(configuration)
        # TODO there might also be a validation/check on the target input

        super().__init__(
            configuration=configuration,  # type:ignore
            target=target,  # type:ignore
        )

    def get_generator(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ) -> Targeted[T]:
        """Get the function to sample batches via AdvancedManufacturing's generator.

        Args:
            configuration: helps to set up specific application of AdvancedManufacturing.
            target: context or condition for the generation.

        Returns:
            callable with target generating a batch of items.
        """
        logger.info("ensure artifacts for the application are present.")
        self.local_artifacts = configuration.ensure_artifacts()
        implementation: Generator = configuration.get_conditional_generator(  # type: ignore
            self.local_artifacts
        )
        return implementation.generate_samples


@ApplicationsRegistry.register_algorithm_application(AdvancedManufacturing)
class CatalystGenerator(AlgorithmConfiguration[SmallMolecule, float]):
    """Configuration to generate catalysts with a desired binding energy."""

    algorithm_type: ClassVar[str] = "controlled_sampling"
    domain: ClassVar[str] = "materials"
    algorithm_version: str = "v0"

    number_of_points: int = field(
        default=32,
        metadata=dict(
            description="Number of points to sample with the Gaussian Process."
        ),
    )
    number_of_steps: int = field(
        default=50,
        metadata=dict(
            description="Number of optimization steps in the Gaussian Process optimization."
        ),
    )
    generated_length: int = field(
        default=100,
        metadata=dict(
            description="Maximum length in tokens of the generated molcules (relates to the SMILES length)."
        ),
    )
    primer_smiles: str = field(
        default="",
        metadata=dict(
            description="Primer molecule to initiate the sampling in SMILES format. Defaults to no primer."
        ),
    )

    def get_target_description(self) -> Dict[str, str]:
        """Get description of the target for generation.

        Returns:
            target description.
        """
        return {
            "title": "Target energy",
            "description": "Binding energy target for the catalysts generated.",
            "type": "number",
        }

    def get_conditional_generator(self, resources_path: str) -> Generator:
        """Instantiate the actual generator implementation.

        Args:
            resources_path: local path to model files.

        Returns:
            instance with :meth:`generate_batch<gt4sd.algorithms.controlled_sampling.advanced_manufacturing.implementation.core.Generator.generate_samples>` method for targeted generation.
        """
        return NCCRCatalystGenerator(
            resources_path=resources_path,
            generated_length=self.generated_length,
            number_of_points=self.number_of_points,
            number_of_steps=self.number_of_steps,
            primer_smiles=self.primer_smiles,
        )

    @classmethod
    def get_filepath_mappings_for_training_pipeline_arguments(
        cls, training_pipeline_arguments: TrainingPipelineArguments
    ) -> Dict[str, str]:
        """Ger filepath mappings for the given training pipeline arguments.

        Args:
            training_pipeline_arguments: training pipeline arguments.

        Returns:
            a mapping between artifacts' files and training pipeline's output files.
        """
        if isinstance(training_pipeline_arguments, GranularSavingArguments):
            return {
                "epoch=199-step=5799.ckpt": training_pipeline_arguments.model_path,
            }
        else:
            return super().get_filepath_mappings_for_training_pipeline_arguments(
                training_pipeline_arguments
            )



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Module initialization for gt4sd traning pipelines."""

import json
import logging
from typing import Any, Dict

import pkg_resources

from ..cli.load_arguments_from_dataclass import extract_fields_from_class
from .diffusion.core import (
    DiffusionDataArguments,
    DiffusionForVisionTrainingPipeline,
    DiffusionModelArguments,
    DiffusionSavingArguments,
    DiffusionTrainingArguments,
)
from .guacamol_baselines.core import GuacaMolDataArguments, GuacaMolSavingArguments
from .guacamol_baselines.smiles_lstm.core import (
    GuacaMolLSTMModelArguments,
    GuacaMolLSTMTrainingArguments,
    GuacaMolLSTMTrainingPipeline,
)
from .moses.core import MosesDataArguments, MosesSavingArguments
from .moses.organ.core import (
    MosesOrganModelArguments,
    MosesOrganTrainingArguments,
    MosesOrganTrainingPipeline,
)
from .moses.vae.core import (
    MosesVAEModelArguments,
    MosesVAETrainingArguments,
    MosesVAETrainingPipeline,
)
from .paccmann.core import (
    PaccMannDataArguments,
    PaccMannSavingArguments,
    PaccMannTrainingArguments,
)
from .paccmann.vae.core import PaccMannVAEModelArguments, PaccMannVAETrainingPipeline
from .pytorch_lightning.core import PytorchLightningTrainingArguments
from .pytorch_lightning.gflownet.core import (
    GFlowNetDataArguments,
    GFlowNetModelArguments,
    GFlowNetPytorchLightningTrainingArguments,
    GFlowNetSavingArguments,
    GFlowNetTrainingPipeline,
)
from .pytorch_lightning.granular.core import (
    GranularDataArguments,
    GranularModelArguments,
    GranularPytorchLightningTrainingArguments,
    GranularSavingArguments,
    GranularTrainingPipeline,
)
from .pytorch_lightning.language_modeling.core import (
    LanguageModelingDataArguments,
    LanguageModelingModelArguments,
    LanguageModelingSavingArguments,
    LanguageModelingTrainingPipeline,
)
from .regression_transformer.core import (
    RegressionTransformerDataArguments,
    RegressionTransformerSavingArguments,
    RegressionTransformerTrainingArguments,
)
from .regression_transformer.implementation import (
    RegressionTransformerModelArguments,
    RegressionTransformerTrainingPipeline,
)
from .torchdrug.core import (
    TorchDrugDataArguments,
    TorchDrugSavingArguments,
    TorchDrugTrainingArguments,
)
from .torchdrug.gcpn.core import (
    TorchDrugGCPNModelArguments,
    TorchDrugGCPNTrainingPipeline,
)
from .torchdrug.graphaf.core import (
    TorchDrugGraphAFModelArguments,
    TorchDrugGraphAFTrainingPipeline,
)

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

TRAINING_PIPELINE_NAME_METADATA_MAPPING = {
    "mock_training_pipeline": "mock_training_pipeline.json",
    "Terminator training": "terminator_training.json",
}

TRAINING_PIPELINE_ARGUMENTS_MAPPING = {
    "language-modeling-trainer": (
        PytorchLightningTrainingArguments,
        LanguageModelingDataArguments,
        LanguageModelingModelArguments,
    ),
    "paccmann-vae-trainer": (
        PaccMannTrainingArguments,
        PaccMannDataArguments,
        PaccMannVAEModelArguments,
    ),
    "torchdrug-gcpn-trainer": (
        TorchDrugTrainingArguments,
        TorchDrugDataArguments,
        TorchDrugGCPNModelArguments,
    ),
    "torchdrug-graphaf-trainer": (
        TorchDrugTrainingArguments,
        TorchDrugDataArguments,
        TorchDrugGraphAFModelArguments,
    ),
    "granular-trainer": (
        GranularPytorchLightningTrainingArguments,
        GranularDataArguments,
        GranularModelArguments,
    ),
    "guacamol-lstm-trainer": (
        GuacaMolLSTMModelArguments,
        GuacaMolLSTMTrainingArguments,
        GuacaMolDataArguments,
    ),
    "moses-vae-trainer": (
        MosesVAETrainingArguments,
        MosesVAEModelArguments,
        MosesDataArguments,
    ),
    "moses-organ-trainer": (
        MosesOrganTrainingArguments,
        MosesOrganModelArguments,
        MosesDataArguments,
    ),
    "regression-transformer-trainer": (
        RegressionTransformerTrainingArguments,
        RegressionTransformerDataArguments,
        RegressionTransformerModelArguments,
    ),
    "diffusion-trainer": (
        DiffusionTrainingArguments,
        DiffusionDataArguments,
        DiffusionModelArguments,
    ),
    "gflownet-trainer": (
        GFlowNetPytorchLightningTrainingArguments,
        GFlowNetDataArguments,
        GFlowNetModelArguments,
    ),
}

TRAINING_PIPELINE_MAPPING = {
    "language-modeling-trainer": LanguageModelingTrainingPipeline,
    "paccmann-vae-trainer": PaccMannVAETrainingPipeline,
    "torchdrug-gcpn-trainer": TorchDrugGCPNTrainingPipeline,
    "torchdrug-graphaf-trainer": TorchDrugGraphAFTrainingPipeline,
    "granular-trainer": GranularTrainingPipeline,
    "guacamol-lstm-trainer": GuacaMolLSTMTrainingPipeline,
    "moses-organ-trainer": MosesOrganTrainingPipeline,
    "moses-vae-trainer": MosesVAETrainingPipeline,
    "regression-transformer-trainer": RegressionTransformerTrainingPipeline,
    "diffusion-trainer": DiffusionForVisionTrainingPipeline,
    "gflownet-trainer": GFlowNetTrainingPipeline,
}

TRAINING_PIPELINE_ARGUMENTS_FOR_MODEL_SAVING = {
    "paccmann-vae-trainer": PaccMannSavingArguments,
    "torchdrug-gcpn-trainer": TorchDrugSavingArguments,
    "torchdrug-graphaf-trainer": TorchDrugSavingArguments,
    "granular-trainer": GranularSavingArguments,
    "language-modeling-trainer": LanguageModelingSavingArguments,
    "guacamol-lstm-trainer": GuacaMolSavingArguments,
    "moses-organ-trainer": MosesSavingArguments,
    "moses-vae-trainer": MosesSavingArguments,
    "regression-transformer-trainer": RegressionTransformerSavingArguments,
    "diffusion-trainer": DiffusionSavingArguments,
    "gflownet-trainer": GFlowNetSavingArguments,
}


def training_pipeline_name_to_metadata(name: str) -> Dict[str, Any]:
    """Retrieve training pipeline metadata from the name.

    Args:
        name: name of the pipeline.

    Returns:
        dictionary describing the parameters of the pipeline. If the pipeline is not found, no metadata (a.k.a., an empty dictionary is returned).
    """
    metadata: Dict[str, Any] = {"training_pipeline": name, "parameters": {}}
    if name in TRAINING_PIPELINE_NAME_METADATA_MAPPING:
        try:
            with open(
                pkg_resources.resource_filename(
                    "gt4sd",
                    f"training_pipelines/{TRAINING_PIPELINE_NAME_METADATA_MAPPING[name]}",
                ),
                "rt",
            ) as fp:
                metadata["parameters"] = json.load(fp)
        except Exception:
            logger.exception(
                f'training pipeline "{name}" metadata fetching failed, returning an empty metadata dictionary'
            )

    elif name in TRAINING_PIPELINE_ARGUMENTS_MAPPING:

        for training_argument_class in TRAINING_PIPELINE_ARGUMENTS_MAPPING[name]:
            field_types = extract_fields_from_class(training_argument_class)
            metadata["parameters"].update(field_types)

    else:
        logger.warning(
            f'training pipeline "{name}" metadata not found, returning an empty metadata dictionary'
        )
    metadata["description"] = metadata["parameters"].pop(
        "description", "A training pipeline."
    )
    return metadata



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""TorchDrug GraphAF trainer unit tests."""

import shutil
import tempfile
from typing import Any, Dict, cast

import pkg_resources

from gt4sd.training_pipelines import (
    TRAINING_PIPELINE_MAPPING,
    TorchDrugGCPNTrainingPipeline,
)

template_config = {
    "model_args": {
        "hidden_dims": "[128, 128]",
        "batch_norm": True,
        "short_cut": True,
        "concat_hidden": True,
        "readout": "mean",
        "hidden_dim_mlp": 128,
        "agent_update_interval": 16,
        "gamma": 0.95,
        "reward_temperature": 1.2,
        "criterion": "{'nll': 1}",
    },
    "dataset_args": {
        "dataset_name": "freesolv",
        "lazy": True,
        "no_kekulization": False,
    },
    "training_args": {
        "model_path": "/tmp/torchdrug-graphaf",
        "training_name": "torchdrug-graphaf-test",
        "epochs": 1,
        "batch_size": 4,
        "learning_rate": 0.0005,
        "log_interval": 2,
        "gradient_interval": 2,
    },
}


def test_train():

    pipeline = TRAINING_PIPELINE_MAPPING.get("torchdrug-graphaf-trainer")

    assert pipeline is not None

    TEMPORARY_DIRECTORY = tempfile.mkdtemp()

    test_pipeline = cast(TorchDrugGCPNTrainingPipeline, pipeline())

    config: Dict[str, Any] = template_config.copy()
    config["training_args"]["model_path"] = TEMPORARY_DIRECTORY
    config["training_args"]["dataset_path"] = TEMPORARY_DIRECTORY

    test_pipeline.train(**config)
    shutil.rmtree(TEMPORARY_DIRECTORY)

    # Now test with a custom dataset
    file_path = pkg_resources.resource_filename(
        "gt4sd",
        "training_pipelines/tests/molecules.csv",
    )
    config["dataset_args"]["dataset_name"] = "custom"
    config["dataset_args"]["file_path"] = file_path
    config["dataset_args"]["smiles_field"] = "smiles"
    config["dataset_args"]["target_field"] = "qed"

    test_pipeline.train(**config)
    shutil.rmtree(TEMPORARY_DIRECTORY)

    # Test the property optimization
    """
    Disabled for now due to multiple downstream torchdrug issues, most importantly:
    - https://github.com/DeepGraphLearning/torchdrug/issues/83
    """
    # config["dataset_args"]["target_field"] = 'qed'
    # config['training_args']['task'] = 'qed'
    # config['data_args']['node_feature'] = 'symbol'
    # config['model_args']['criterion'] = "{'ppo': 1}"

    # test_pipeline.train(**config)
    # shutil.rmtree(TEMPORARY_DIRECTORY)



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""CLaSS Algorithm: PAG and CogMol applications."""

import logging
from dataclasses import field
from typing import Any, Callable, ClassVar, Dict, Iterable, Optional, TypeVar, Union

from ....extras import EXTRAS_ENABLED
from ...core import AlgorithmConfiguration, GeneratorAlgorithm  # type: ignore
from ...registry import ApplicationsRegistry  # type: ignore

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

if EXTRAS_ENABLED:
    from cog.core import CogMolGenerator
    from cog.sample_pipeline import CogMolFiles, read_artifacts_config
    from pag.core import PAGGenerator
    from pag.sample_pipeline import PAGFiles

    T = TypeVar("T")
    S = TypeVar("S")
    Targeted = Callable[[T], Iterable[Any]]
    Untargeted = Callable[[], Iterable[Any]]

    class CLaSS(GeneratorAlgorithm[S, T]):
        """Controlled Latent attribute Space Sampling (CLaSS) Algorithm."""

        def __init__(
            self,
            configuration: AlgorithmConfiguration[S, T],
            target: Optional[T] = None,
        ):
            """Instantiate CLaSS ready to generate items.

            Args:
                configuration: domain and application
                    specification, defining types and validations.
                target: Optional, in this inistance we will convert to a string.

            Example:
                An example for using the CogMol application with this Algorithm::

                    # target protein
                    MPRO = "SGFRKMAFPSGKVEGCMVQVTCGTTTLNGLWLDDVVYCPRHVICTSEDMLNPNYEDLLIRKSNHNFLVQAGNVQLRVIGHSMQNCVLKLKVDTANPKTPKYKFVRIQPGQTFSVLACYNGSPSGVYQCAMRPNFTIKGSFLNGSCGSVGFNIDYDCVSFCYMHHMELPTGVHAGTDLEGNFYGPFVDRQTAQAAGTDTTITVNVLAWLYAAVINGDRWFLNRFTTTLNDFNLVAMKYNYEPLTQDHVDILGPLSAQTGIAVLDMCASLKELLQNGMNGRTILGSALLEDEFTPFDVVRQCSGVTFQ"
                    configuration = CogMol()
                    algorithm = CLaSS(configuration=configuration, target=MPRO)
                    items = list(algorithm.sample(1))
                    print(items)

                We can also use the PAG application similarly::

                    configuration = PAG()
                    algorithm = CLaSS(configuration=configuration)
                    items = list(algorithm.sample(1))
                    print(items)
            """

            configuration = self.validate_configuration(configuration)
            # TODO there might also be a validation/check on the target input

            super().__init__(
                configuration=configuration,
                target=target,  # type:ignore
            )

        def get_generator(
            self,
            configuration: AlgorithmConfiguration[S, T],
            target: Optional[T],
        ) -> Union[Untargeted, Targeted[T]]:
            """Get the function to sample from generator.

            Args:
                configuration: helps to set up the application.
                target: target to generate molecules against.

            Returns:
                callable generating a list of molecules.
            """
            logger.info("ensure artifacts for the application are present.")
            self.local_artifacts = configuration.ensure_artifacts()
            implementation = configuration.get_class_instance(  # type: ignore
                resources_path=self.local_artifacts, target=target
            )
            return implementation.sample_accepted

        def validate_configuration(
            self, configuration: AlgorithmConfiguration
        ) -> AlgorithmConfiguration:
            # TODO raise InvalidAlgorithmConfiguration
            assert isinstance(configuration, AlgorithmConfiguration)
            return configuration

    @ApplicationsRegistry.register_algorithm_application(CLaSS)
    class CogMol(AlgorithmConfiguration[str, str]):
        """Configuration for CogMol: Target-Specific and Selective Drug Design."""

        algorithm_type: ClassVar[str] = "controlled_sampling"
        domain: ClassVar[str] = "materials"
        algorithm_version: str = "v0"

        samples_per_round: int = field(
            default=200,
            metadata=dict(
                description="Number of generated samples for acceptance/rejection per round."
            ),
        )
        max_length: int = field(
            default=100,
            metadata=dict(description="Maximal number of tokens in generated samples."),
        )
        temperature: float = field(
            default=1.0,
            metadata=dict(description="Temperature of softmax."),
        )
        num_proteins_selectivity: int = field(
            default=10,
            metadata=dict(
                description="Number of random samples for measuring off target selectivity for rejection."
            ),
        )

        def get_target_description(self) -> Dict[str, str]:
            """Get description of the target for generation.
            Returns:
                target description.
            """
            return {
                "title": "Protein",
                "description": "Primary structure of the target protein as sequence of amino acid characters.",
                "type": "string",
            }

        def get_class_instance(self, resources_path: str, target: str):
            try:
                config = read_artifacts_config(resources_path)
                bindingdb_date = config["cogmol version information"]["bindingdb_date"]
            except KeyError:
                bindingdb_date = None

            return CogMolGenerator(
                protein_sequence=target,
                model_files=CogMolFiles.from_directory_with_config(resources_path),
                n_samples_per_round=self.samples_per_round,
                device="cpu",
                num_proteins_selectivity=self.num_proteins_selectivity,
                temp=self.temperature,
                max_len=self.max_length,
                bindingdb_date=bindingdb_date,
            )

    @ApplicationsRegistry.register_algorithm_application(CLaSS)
    class PAG(AlgorithmConfiguration[str, str]):
        """Configuration for photoacid generator (PAG) design."""

        algorithm_type: ClassVar[str] = "controlled_sampling"
        domain: ClassVar[str] = "materials"
        algorithm_version: str = "v0"

        samples_per_round: int = field(
            default=200,
            metadata=dict(
                description="Number of generated samples for acceptance/rejection per round."
            ),
        )
        max_length: int = field(
            default=100,
            metadata=dict(description="Maximal number of tokens in generated samples."),
        )
        temperature: float = field(
            default=1.0,
            metadata=dict(description="Temperature of softmax."),
        )

        def get_target_description(self) -> None:
            """Untargeted sampling. Always returns None.

            Returns:
                None
            """
            return None

        def get_class_instance(self, resources_path: str, target: Optional[T] = None):
            if target is not None:
                raise NotImplementedError

            return PAGGenerator(
                model_files=PAGFiles.from_directory_with_config(resources_path),
                n_samples_per_round=self.samples_per_round,
                device="cpu",
                temp=self.temperature,
                max_len=self.max_length,
            )

else:
    logger.warning("install cogmol-inference extras to use CLaSS")



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from argparse import Namespace

import numpy as np
import pytest
import pytorch_lightning as pl

from gt4sd.frameworks.gflownet.dataloader.data_module import GFlowNetDataModule
from gt4sd.frameworks.gflownet.envs.graph_building_env import GraphBuildingEnv
from gt4sd.frameworks.gflownet.envs.mol_building_env import MolBuildingEnvContext
from gt4sd.frameworks.gflownet.loss import ALGORITHM_FACTORY
from gt4sd.frameworks.gflownet.ml.models import MODEL_FACTORY
from gt4sd.frameworks.gflownet.ml.module import GFlowNetModule
from gt4sd.frameworks.gflownet.tests.qm9 import QM9Dataset, QM9GapTask

configuration = {
    "bootstrap_own_reward": False,
    "learning_rate": 1e-4,
    "global_batch_size": 16,
    "num_emb": 128,
    "num_layers": 4,
    "tb_epsilon": 1e-10,
    "illegal_action_logreward": -50,
    "reward_loss_multiplier": 1,
    "temperature_sample_dist": "uniform",
    "temperature_dist_params": "(.5, 32)",
    "weight_decay": 1e-8,
    "num_data_loader_workers": 8,
    "momentum": 0.9,
    "adam_eps": 1e-8,
    "lr_decay": 20000,
    "z_lr_decay": 20000,
    "clip_grad_type": "norm",
    "clip_grad_param": 10,
    "random_action_prob": 0.001,
    "sampling_tau": 0.0,
    "max_nodes": 9,
    "num_offline": 10,
    "sampling_iterator": True,
    "ratio": 0.9,
    "development": True,
    "epoch": 1,
    "batch_size": 64,
    "num_workers": 0,
    "lr": 0.0001,
    "algorithm": "trajectory_balance",
    "dataset": "qm9",
    "dataset_path": "./data/qm9.h5",
    "model": "graph_transformer_gfn",
    "sampling_model": "graph_transformer_gfn",
    "task": "qm9",
    "device": "cpu",
    "seed": 124,
    "test_output_path": "logs",
}
configuration["rng"] = np.random.default_rng(configuration["seed"])  # type: ignore


def test_gfn_env():
    environment = GraphBuildingEnv()
    context = MolBuildingEnvContext()

    assert isinstance(environment, GraphBuildingEnv)
    assert isinstance(context, MolBuildingEnvContext)


@pytest.mark.parametrize("model_name", ["graph_transformer_gfn"])
def test_gfn_model(model_name):
    context = MolBuildingEnvContext()
    model = MODEL_FACTORY[model_name](
        configuration,
        context,
    )
    assert isinstance(model, MODEL_FACTORY[model_name])


@pytest.mark.skip(reason="we need to add support for dataset buckets")
def test_gfn():
    """test basic GFN training on QM9."""

    dataset = QM9Dataset(configuration["dataset_path"], target="gap")  # type: ignore
    environment = GraphBuildingEnv()
    context = MolBuildingEnvContext()

    algorithm = ALGORITHM_FACTORY[configuration["algorithm"]](  # type: ignore
        configuration=configuration,
        environment=environment,
        context=context,
    )
    model = MODEL_FACTORY[configuration["model"]](  # type: ignore
        configuration=configuration,
        context=context,
    )

    task = QM9GapTask(
        configuration=configuration,
        dataset=dataset,
    )

    dm = GFlowNetDataModule(
        configuration=configuration,
        dataset=dataset,
        environment=environment,
        context=context,
        task=task,
        algorithm=algorithm,
        model=model,
    )
    dm.prepare_data()

    module = GFlowNetModule(
        configuration=configuration,
        dataset=dataset,
        environment=environment,
        context=context,
        task=task,
        algorithm=algorithm,
        model=model,
    )

    trainer = pl.Trainer.from_argparse_args(
        Namespace(**configuration),
        profiler="simple",
        auto_lr_find=True,
        log_every_n_steps=50,
        max_epochs=1,
        flush_logs_every_n_steps=100,
        fast_dev_run=True,
        accelerator="ddp",
    )
    trainer.fit(module, dm)



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Data module for gflownet."""

import logging
from typing import Any, Dict, Optional

import sentencepiece as _sentencepiece
import numpy as np
import pytorch_lightning as pl
import torch.nn as nn
from torch.utils.data import DataLoader  # , Subset, random_split

from ..dataloader.dataset import GFlowNetDataset, GFlowNetTask
from ..envs.graph_building_env import GraphBuildingEnv, GraphBuildingEnvContext
from ..loss.trajectory_balance import TrajectoryBalance
from ..ml.models import MODEL_FACTORY
from .sampler import SamplingIterator

# sentencepiece has to be loaded before lightning to avoid segfaults
_sentencepiece

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class GFlowNetDataModule(pl.LightningDataModule):
    """Data module from gflownet."""

    def __init__(
        self,
        configuration: Dict[str, Any],
        dataset: GFlowNetDataset,
        environment: GraphBuildingEnv,
        context: GraphBuildingEnvContext,
        task: GFlowNetTask,
        algorithm: TrajectoryBalance,
        model: Optional[nn.Module] = None,
    ) -> None:
        """Construct GFlowNetDataModule.

        The module assumes a model and algorithm factory/registry.
        The user should provide a dataset, environment, context for the environment, and task.

        Args:
            configuration: configuration dictionary.
            dataset: dataset.
            environment: environment for graph building.
            context: context environment.
            task: generic task.
            algorithm: loss function.
            model: model used to generate data with the sampling iterator.
                It can be a custom model or the same as the one used in the algorithm.
        """
        super().__init__()
        self.hps = configuration

        # if model is given
        if model:
            self.sampling_model = model
        else:
            self.sampling_model = MODEL_FACTORY[self.hps["sampling_model"]](
                self.hps, context
            )
        self.algo = algorithm
        self.env = environment
        self.ctx = context
        self.dataset = dataset
        self.task = task

        self.sampling_iterator = self.hps["sampling_iterator"]
        self.batch_size = self.hps["batch_size"]
        self.num_workers = self.hps["num_workers"]
        self.device = self.hps["device"]
        self.rng = self.hps["rng"]
        self.ratio = self.hps["ratio"]
        self.mb_size = self.hps["global_batch_size"]

    def prepare_data(self) -> None:
        """Prepare training and test dataset."""
        self.train_dataset = self.dataset
        self.val_dataset = self.dataset
        self.test_dataset = self.dataset

    def setup(self, stage: Optional[str]) -> None:  # type:ignore
        """Setup the data module.

        Args:
            stage: stage considered. Defaults to None.
        """

        ll = self.dataset.get_len()
        ixs = np.arange(ll)
        self.rng.shuffle(ixs)
        thresh = int(np.floor(self.ratio * ll))

        self.ix_train = ixs[: int(0.9 * thresh)]
        self.ix_val = ixs[int(0.9 * thresh) : thresh]
        self.ix_test = ixs[thresh:]

        if stage == "fit" or stage is None:
            self.train_dataset.set_indexes(self.ix_train)
            self.val_dataset.set_indexes(self.ix_val)
        if stage == "test" or stage is None:
            self.test_dataset.set_indexes(self.ix_test)
        if stage == "predict" or stage is None:
            self.test_dataset.set_indexes(self.ix_test)

        logger.info(
            f"number of data points used for training: {len(self.train_dataset)}"
        )
        logger.info(f"number of data points used for testing: {len(self.test_dataset)}")
        logger.info(
            f"testing proportion: {len(self.test_dataset) / (len(self.test_dataset) + len(self.train_dataset))}"
        )

    def train_dataloader(self) -> DataLoader:
        """Get a data loader for training.

        Returns:
            a training data loader.
        """
        if self.sampling_iterator:
            iterator = SamplingIterator(
                self.train_dataset,
                self.sampling_model,
                self.mb_size * 2,
                self.ctx,
                self.algo,
                self.task,
                device=self.device,
            )
            batch_size = None
        else:
            iterator = self.train_dataset  # type: ignore
            batch_size = self.batch_size

        return DataLoader(
            iterator,
            batch_size=batch_size,
            num_workers=self.num_workers,
            persistent_workers=self.num_workers > 0,
        )

    def val_dataloader(self) -> DataLoader:
        """Get a data loader for validation.

        Returns:
            a validation data loader.
        """
        if self.sampling_iterator:
            iterator = SamplingIterator(
                self.val_dataset,
                self.sampling_model,
                self.mb_size,
                self.ctx,
                self.algo,
                self.task,
                ratio=1,
                stream=False,
                device=self.device,
            )
            batch_size = None
        else:
            iterator = self.val_dataset  # type: ignore
            batch_size = self.batch_size

        return DataLoader(
            iterator,
            batch_size=batch_size,
            num_workers=self.num_workers,
            persistent_workers=self.num_workers > 0,
        )

    def test_dataloader(self) -> DataLoader:
        """Get a data loader for testing.

        Returns:
            a testing data loader.
        """
        return DataLoader(
            self.test_dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            pin_memory=False,
        )

    def predict_dataloader(self) -> DataLoader:
        """Get a data loader for prediction.

        Returns:
            a prediction data loader.
        """
        return DataLoader(
            self.test_dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            pin_memory=False,
        )



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Train module implementation."""

import logging
from argparse import Namespace
from typing import Any, Dict

import sentencepiece as _sentencepiece
import pytorch_lightning as pl
from pytorch_lightning.callbacks import ModelCheckpoint
from pytorch_lightning.loggers import TensorBoardLogger

from ..arg_parser.parser import parse_arguments_from_config

# from ..dataloader import build_dataset
from ..dataloader.data_module import GFlowNetDataModule
from ..dataloader.dataset import GFlowNetDataset, GFlowNetTask

# from ..envs import build_env_context
from ..envs.graph_building_env import GraphBuildingEnv, GraphBuildingEnvContext
from ..loss import ALGORITHM_FACTORY
from ..ml.models import MODEL_FACTORY
from ..ml.module import GFlowNetModule

# from ..train import build_task

# sentencepiece has to be loaded before lightning to avoid segfaults
_sentencepiece

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def train_gflownet(
    configuration: Dict[str, Any],
    dataset: GFlowNetDataset,
    environment: GraphBuildingEnv,
    context: GraphBuildingEnvContext,
    task: GFlowNetTask,
) -> None:
    """Train a gflownet given a configuration, a dataset and a task.
    The default enviroment and context are compatible with small molecules.

    Args:
        configuration: a configuration dictionary.
        dataset: a dataset compatible with lightning.
        environment: an environment specifying the state space.
        context: an environment context specifying how to combine states.
        task: a task specifying the reward structure.
    """

    arguments = Namespace(**configuration)

    if arguments.algorithm in ALGORITHM_FACTORY:
        algorithm = ALGORITHM_FACTORY[getattr(arguments, "algorithm")](
            configuration=configuration,
            environment=environment,
            context=context,
        )
    else:
        raise ValueError(f"Algorithm {arguments.algorithm} not supported.")

    if arguments.model in MODEL_FACTORY:
        model = MODEL_FACTORY[getattr(arguments, "model")](
            configuration=configuration,
            context=context,
        )
    else:
        raise ValueError(f"Model {arguments.model} not supported.")

    dm = GFlowNetDataModule(
        configuration=configuration,
        dataset=dataset,
        environment=environment,
        context=context,
        task=task,
        algorithm=algorithm,
        model=model,
    )
    dm.prepare_data()

    module = GFlowNetModule(
        configuration=configuration,
        dataset=dataset,
        environment=environment,
        context=context,
        task=task,
        algorithm=algorithm,
        model=model,
    )

    tensorboard_logger = TensorBoardLogger(
        "logs", name=getattr(arguments, "basename", "default")
    )
    checkpoint_callback = ModelCheckpoint(
        save_top_k=-1,
    )
    trainer = pl.Trainer.from_argparse_args(
        arguments,
        profiler="simple",
        logger=tensorboard_logger,
        auto_lr_find=True,
        log_every_n_steps=getattr(arguments, "trainer_log_every_n_steps", 50),
        callbacks=[checkpoint_callback],
        max_epochs=getattr(arguments, "epoch", 10),
        check_val_every_n_epoch=getattr(arguments, "checkpoint_every_n_val_epochs", 5),
        fast_dev_run=getattr(arguments, "development_mode", False),
        accelerator=getattr(arguments, "distributed_training_strategy", "ddp"),
    )
    trainer.fit(module, dm)


def train_gflownet_main(
    configuration: Dict[str, Any],
    dataset: GFlowNetDataset,
    environment: GraphBuildingEnv,
    context: GraphBuildingEnvContext,
    task: GFlowNetTask,
) -> None:
    """Train a gflownet module parsing arguments from config and standard input."""

    # add user configuration
    configuration.update(vars(parse_arguments_from_config()))
    # train gflownet
    train_gflownet(
        configuration=configuration,
        dataset=dataset,
        environment=environment,
        context=context,
        task=task,
    )



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import logging
from dataclasses import field
from typing import Any, Callable, ClassVar, Dict, Iterable, Optional, TypeVar

from ....domains.materials import SMILES, validate_molecules
from ....exceptions import InvalidItem
from ...core import AlgorithmConfiguration, GeneratorAlgorithm
from ...registry import ApplicationsRegistry
from .implementation import ReinventConditionalGenerator

T = TypeVar("T", bound=Any)
S = TypeVar("S", bound=Any)
Targeted = Callable[[T], Iterable[Any]]

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class Reinvent(GeneratorAlgorithm[S, T]):
    """Reinvent sample generation algorithm."""

    def __init__(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ):
        """Instantiate Reinvent ready to generate samples.

        Args:
            configuration: domain and application
                specification defining parameters, types and validations.
            target: a target for which to generate items.

        Example:
            An example for predicting topics for a given text::

                config = ReinventGenerator()
                algorithm = Reinvent(configuration=config, target="CCO")
                items = list(algorithm.sample(1))
                print(items)
        """

        configuration = self.validate_configuration(configuration)
        # TODO there might also be a validation/check on the target input

        super().__init__(
            configuration=configuration,  # type:ignore
            target=target,  # type:ignore
        )

    def get_generator(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ) -> Targeted[T]:
        """Get the function to perform the prediction via Reinvent's generator.

        Args:
            configuration: helps to set up specific application of Reinvent.
            target: context or condition for the generation.

        Returns:
            callable with target generating samples.
        """
        logger.info("ensure artifacts for the application are present.")
        self.local_artifacts = configuration.ensure_artifacts()
        implementation: ReinventConditionalGenerator = configuration.get_conditional_generator(  # type: ignore
            self.local_artifacts
        )
        return implementation.generate_samples


@ApplicationsRegistry.register_algorithm_application(Reinvent)
class ReinventGenerator(AlgorithmConfiguration[str, str]):
    """Configuration to generate molecules using the REINVENT algorithm. It generates the molecules minimizing the distances between the scaffolds."""

    algorithm_name: ClassVar[str] = Reinvent.__name__
    algorithm_type: ClassVar[str] = "conditional_generation"
    domain: ClassVar[str] = "materials"
    algorithm_version: str = "v0"

    batch_size: int = field(
        default=20,
        metadata=dict(description=("Number of samples to generate per scaffold")),
    )

    randomize: bool = field(
        default=True,
        metadata=dict(description=("Randomize the scaffolds if set to true")),
    )

    sample_uniquely: bool = field(
        default=True,
        metadata=dict(description=("Generate unique sample sequences if set to true")),
    )
    max_sequence_length: int = field(
        default=256,
        metadata=dict(description=("Maximal length of SMILES sequences")),
    )

    def get_target_description(self) -> Dict[str, str]:
        """Get description of the target for generation.

        Returns:
            target description.
        """
        return {
            "title": "SMILES for sample generation",
            "description": "SMILES considered for the samples generation.",
            "type": "string",
        }

    def get_conditional_generator(
        self, resources_path: str
    ) -> ReinventConditionalGenerator:
        """Instantiate the actual generator implementation.

        Args:
            resources_path: local path to model files.

        Returns:
            instance with :meth:`generate_samples<gt4sd.algorithms.conditional_generation.reinvent.implementation.ReinventConditionalGenerator.generate_samples>` method for targeted generation.
        """
        return ReinventConditionalGenerator(
            resources_path=resources_path,
            batch_size=self.batch_size,
            randomize=self.randomize,
            sample_uniquely=self.sample_uniquely,
            max_sequence_length=self.max_sequence_length,
        )

    def validate_item(self, item: str) -> SMILES:
        """Check that item is a valid SMILES.

        Args:
            item: a generated item that is possibly not valid.

        Raises:
            InvalidItem: in case the item can not be validated.

        Returns:
            the validated SMILES.
        """
        molecules, _ = validate_molecules(smiles_list=[item])

        if molecules[0] is None:
            raise InvalidItem(
                title="InvalidSMILES",
                detail=f'rdkit.Chem.MolFromSmiles returned None for "{item}"',
            )
        return SMILES(item)



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Enzyme optimization."""

import json
import logging
import random
import time
from collections import OrderedDict
from typing import Any, Dict, List, MutableMapping, Optional, Sequence, Tuple, Union

import numpy as np
from joblib import load

from .processing import (
    HuggingFaceTransformerEmbedding,
    StringEmbedding,
    TAPEEmbedding,
    reconstruct_sequence_with_mutation_range,
    sanitize_intervals,
)

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

#: transition matrix representation
TransitionMatrix = MutableMapping[str, MutableMapping[str, float]]
#: transition matrix configuration
TransitionConfiguration = MutableMapping[
    str, Union[MutableMapping[str, float], Sequence[str]]
]

#: supported features
SUPPORTED_FEATURE_SET = set(["substrate", "product", "sequence"])

#: IUPAC code mapping
IUPAC_CODES = OrderedDict(
    [
        ("Ala", "A"),
        ("Asx", "B"),  # Aspartate or Asparagine
        ("Cys", "C"),
        ("Asp", "D"),
        ("Glu", "E"),
        ("Phe", "F"),
        ("Gly", "G"),
        ("His", "H"),
        ("Ile", "I"),
        ("Lys", "K"),
        ("Leu", "L"),
        ("Met", "M"),
        ("Asn", "N"),
        ("Pyl", "O"),  # Pyrrolysin
        ("Pro", "P"),
        ("Gln", "Q"),
        ("Arg", "R"),
        ("Ser", "S"),
        ("Thr", "T"),
        ("Sec", "U"),  # Selenocysteine
        ("Val", "V"),
        ("Trp", "W"),
        ("Xaa", "X"),  # Any AA
        ("Tyr", "Y"),
        ("Glx", "Z"),  # Glutamate or Glutamine
    ]
)
#: IUPAC character set
IUPAC_CHARACTER_SET = set(IUPAC_CODES.values())
#: IUPAC uniform mutation mapping, we exclude 'X' from the mapping values because it denotes a generic AA
IUPAC_MUTATION_MAPPING: TransitionConfiguration = {
    iupac_character: sorted(list(IUPAC_CHARACTER_SET - {iupac_character, "X"}))
    for iupac_character in IUPAC_CHARACTER_SET
}


class Mutations:
    """Mutations definition class."""

    def __init__(self, transition_configuration: TransitionConfiguration) -> None:
        """Generate the mutation given the configuration for the transitions.

        Args:
            transition_configuration: transition configuration.
        """
        self.transition_matrix = Mutations.transition_configuration_to_matrix(
            transition_configuration
        )

    @staticmethod
    def transition_configuration_to_matrix(
        transition_configuration: TransitionConfiguration,
    ) -> TransitionMatrix:
        """Transform a configuration into a valid transition matrix.

        Args:
            transition_configuration: transition configuration.

        Returns:
            a transition matrix.
        """
        transition_matrix: TransitionMatrix = dict()
        for transition_source, transition_targets in transition_configuration.items():
            if isinstance(transition_targets, dict):
                total = float(sum(transition_targets.values()))
                transition_matrix[transition_source] = {
                    transition_target: transtion_element / total
                    for transition_target, transtion_element in transition_targets.items()
                }
            else:
                transition_matrix[transition_source] = {
                    transition_target: 1 / len(transition_targets)
                    for transition_target in transition_targets
                }
        return transition_matrix

    @staticmethod
    def from_json(filepath: str) -> "Mutations":
        """Parse the mutation from a JSON containing the transition configuration.

        Returns:
            the mutations object.
        """
        with open(filepath) as fp:
            return Mutations(json.load(fp))

    def mutate(self, source: str) -> str:
        """Mutate a source string.

        Args:
            source: source string.

        Returns:
            the mutated target.
        """
        targets, probabilities = zip(*self.transition_matrix[source].items())
        return np.random.choice(targets, size=1, p=probabilities).item()


class AASequence:
    def __init__(
        self, sequence: str, mutations: Mutations = Mutations(IUPAC_MUTATION_MAPPING)
    ) -> None:
        """Initialize an AA sequence representation.

        Args:
            sequence: AA sequence.
            mutations: mutations definition. Defaults to uniform sampling of IUPAC AAs.
        """
        self.sequence = sequence
        self.sequence_length = len(sequence)
        self.mutations = mutations

    def mutate(self, maximum_number_of_mutations: int) -> str:
        """Mutate the sequence in multiple positions.

        Args:
            maximum_number_of_mutations: maximum number of mutations.

        Returns:
            the mutated sequence.
        """
        if maximum_number_of_mutations > self.sequence_length:
            logger.warning(
                f"resetting maximum number of mutations ({maximum_number_of_mutations}), since it is higher than sequence length: {self.sequence_length}"
            )
            maximum_number_of_mutations = self.sequence_length
        if maximum_number_of_mutations < 1:
            logger.warning(
                f"maximum number of mutations can't be lower than 1 ({maximum_number_of_mutations}), resetting to 1"
            )
            maximum_number_of_mutations = 1
        number_of_mutations = random.randint(1, maximum_number_of_mutations)
        positions = sorted(
            random.sample(range(self.sequence_length), number_of_mutations)
        )
        mutated_sequence = ""
        start_position = -1
        for position in positions:
            mutated_sequence += self.sequence[(start_position + 1) : position]
            mutated_sequence += self.mutations.mutate(self.sequence[position])
            start_position = position
        mutated_sequence += self.sequence[(start_position + 1) :]
        return mutated_sequence


class EnzymeOptimizer:
    """Optimize an enzyme to catalyze a reaction from substrate to product."""

    def __init__(
        self,
        scorer_filepath: str,
        substrate: str,
        product: str,
        sequence: str,
        protein_embedding: StringEmbedding = TAPEEmbedding(),
        molecule_embedding: StringEmbedding = HuggingFaceTransformerEmbedding(),
        ordering: List[str] = ["substrate", "product", "sequence"],
    ) -> None:
        """Initialize the enzyme designer.

        Args:
            scorer_filepath: pickled scorer filepath.
            substrate: substrate SMILES.
            product: product SMILES.
            sequence: AA sequence representing the enzyme to optimize.
            protein_embedding: protein embedding class. Defaults to TAPE bert-base.
            molecule_embedding: molecule embedding class. Defaults to ChemBERTa version 1.
            ordering: ordering of the features for the scorer. Defaults to ["substrate", "product", "sequence"].

        Raises:
            ValueError: ordering provided is not feasible.

        Example:
            An example optimizing a specific reaction::

                filepath = f"/path/to/model/scoring_model.pkl"
                substrate = "NC1=CC=C(N)C=C1"
                product = "CNC1=CC=C(NC(=O)C2=CC=C(C=C2)C(C)=O)C=C1"
                sequence = (
                    "MSIQIKQSTMVRPAEETPNKSLWLSNIDMILRTPYSHTGAVLIYKQPDNNEDNIHPSSSMYFDANILIEALSKA"
                    "LVPFYPMAGRLKINGDRYEIDCNAEGALFVEAESSHVLEDFGDFRPNDELHRVMVPTCDYSKGISSFPLLMVQLT"
                    "RFRCGGVSIGFAQHHHVCDGMAHFEFNNSWARIAKGLLPALEPVHDRYLHLRPRNPPQIKYSHSQFEPFVPSLPN"
                    "ELLDGKTNKSQTLFILSREQINTLKQKLDLSNNTTRLSTYEVVAAHVWRSVSKARGLSDHEEIKLIMPVDGRSRIN"
                    "NPSLPKGYCGNVVFLAVCTATVGDLSCNPLTDTAGKVQEALKGLDDDYLRSAIDHTESKPGLPVPYMGSPEKTLYPN"
                    "VLVNSWGRIPYQAMDFGWGSPTFFGISNIFYDGQCFLIPSRDGDGSMTLAINLFSSHLSRFKKYFYDF"
                )
                # instantiate the designer
                designer = EnzymeOptimizer(
                    scorer_filepath=filepath, substrate=substrate, product=product, sequence=sequence
                )


                # with this sequence length every steps takes ~5s
                # optimize between positions 150 and 405 allowing for a maximum of 5 mutations.
                results = designer.optimize(
                    number_of_mutations=5, number_of_steps=10, number_of_samples_per_step=8,
                    intervals=[(150, 405)]
                )
                best_score = results[0]["score"]
                best_sequence = results[0]["sequence"]
        """
        if len(set(ordering).intersection(SUPPORTED_FEATURE_SET)) < 3:
            raise ValueError(
                f"ordering={ordering} should contain only the three admissible values: {sorted(list(SUPPORTED_FEATURE_SET))}"
            )
        else:
            self._ordering = ordering
        self.scorer_filepath = scorer_filepath
        self.scorer = load(scorer_filepath)
        self.substrate = substrate
        self.product = product
        self.protein_embedding = protein_embedding
        self.molecule_embedding = molecule_embedding
        self.embedded_vectors = {
            "substrate": self.molecule_embedding.embed_one(self.substrate),
            "product": self.molecule_embedding.embed_one(self.product),
        }
        self.sequence = sequence
        self.sequence_length = len(sequence)

    def score_sequence(self, sequence: str) -> float:
        """Score a given sequence.

        Args:
            sequence: a sequence to score.

        Returns:
            score for the sequence.
        """
        embedded_vectors = {"sequence": self.protein_embedding.embed_one(sequence)}
        embedded_vectors.update(self.embedded_vectors)
        feature_vector = np.concatenate(
            [embedded_vectors[feature] for feature in self._ordering], axis=1
        )
        return self.scorer.predict_proba(feature_vector)[0][1]

    def score_sequences(self, sequences: List[str]) -> List[Dict[str, Any]]:
        """Score a given sequence list.

        Args:
            sequences: a list of sequences to score.

        Returns:
            a list of dictionaries of sequences and related scores.
        """
        number_of_sequences = len(sequences)
        embedded_matrices = {
            "substrate": np.repeat(
                self.embedded_vectors["substrate"], number_of_sequences, axis=0
            ),
            "product": np.repeat(
                self.embedded_vectors["product"], number_of_sequences, axis=0
            ),
        }
        embedded_matrices["sequence"] = self.protein_embedding(sequences)
        feature_vector = np.concatenate(
            [embedded_matrices[feature] for feature in self._ordering], axis=1
        )
        return [
            {"sequence": sequence, "score": score}
            for sequence, score in zip(
                sequences, self.scorer.predict_proba(feature_vector)[:, 1]
            )
        ]

    def optimize(
        self,
        number_of_mutations: int,
        intervals: Optional[List[Tuple[int, int]]] = None,
        number_of_steps: int = 10,
        number_of_samples_per_step: int = 32,
        number_of_sequences: Optional[int] = None,
        seed: int = 42,
        time_budget: Optional[int] = None,
        mutations: Mutations = Mutations(IUPAC_MUTATION_MAPPING),
    ) -> List[Dict[str, Any]]:
        """Optimize the enzyme given a number of mutations and a range.

        If the range limits are not provided the full sequence is optimized, this might be inefficient.
        The sampling is performing by exploring mutations with a slightly smart random sampling.

        Args:
            number_of_mutations: number of allowed mutations.
            intervals: list of ranges in the sequence, zero-based. Defaults to None, a.k.a. use optimize the full sequence.
            number_of_steps: number of optimization steps. Defaults to 100.
            number_of_samples_per_step: number of samples sequences per optimization step. Defaults to 32.
            number_of_sequences: number of optimal seuqence returned. Defaults to None, a.k.a, returns all.
            seed: seed for random number generation. Defaults to 42.
            time_budget: maximum allowed runtime in seconds. Defaults to None, a.k.a., no time limit, running for number_of_steps steps.
            mutations: mutations definition. Defaults to uniform sampling of IUPAC AAs.

        Raises:
            ValueError: in case an invalid range is provided.

        Returns:
            a list of dictionaries containing a candidate optimal sequence and the related score. Sorted from best to worst.
            Note that, when no limit on the returned number of sequences is set, the worst sequence is the original unmutated sequence.
            If the optimization fails, only the original sequence is returned.
        """
        random.seed(seed)

        # check if interval is None. In case it is, take as interval the whole sequence
        if intervals is None:
            intervals = [(0, self.sequence_length)]
        else:
            intervals = sanitize_intervals(
                intervals
            )  # here we merged and sorted the intervals

        # check that the intervals are in the range of the sequence length
        if intervals[-1][1] > self.sequence_length:
            raise ValueError(
                "check provided intervals, at least an interval is larger than the sequence length"
            )

        # create a sequence from based on the intervals
        sequence_from_intervals = "".join(
            [self.sequence[start:end] for start, end in intervals]
        )

        # mutate the sequence from intervals
        aa_sequence_range = AASequence(sequence_from_intervals, mutations=mutations)
        maximum_number_of_mutations = number_of_mutations

        logger.info(
            f"maximum number of mutations for the intervals: {maximum_number_of_mutations}"
        )
        scored_original_sequence = {
            "score": self.score_sequence(self.sequence),
            "sequence": self.sequence,
        }
        original_sequence_score = scored_original_sequence["score"]
        logger.info(f"original sequence score: {original_sequence_score}")
        results: List[Dict[str, Any]] = [scored_original_sequence]
        # slightly smart random sampling
        visited_sequences = set()
        start_time = time.time()
        for step in range(number_of_steps):
            logger.info(f"optimization step={step + 1}")
            mutated_sequences = []

            for _ in range(number_of_samples_per_step):
                mutated_sequence_range = aa_sequence_range.mutate(
                    maximum_number_of_mutations=maximum_number_of_mutations
                )

                mutated_sequence = reconstruct_sequence_with_mutation_range(
                    sequence=self.sequence,
                    mutated_sequence_range=mutated_sequence_range,
                    intervals=intervals,
                )

                # make sure we do not revisit
                if mutated_sequence not in visited_sequences:
                    visited_sequences.add(mutated_sequence)
                    mutated_sequences.append(mutated_sequence)

            # add only mutated sequences that are more optimal than the original
            results += [
                scored_sequence
                for scored_sequence in self.score_sequences(mutated_sequences)
                if scored_sequence["score"] > original_sequence_score
            ]
            logger.info(
                f"best score at step={step + 1}: {max([scored_sequence['score'] for scored_sequence in results])}"
            )
            elapsed_time = int(time.time() - start_time)
            if time_budget is not None:
                if elapsed_time > time_budget:
                    logger.warning(
                        f"used all the given time budget of {time_budget}s, exting optimization loop"
                    )
                    break
        logger.info(
            f"optimization completed visiting {len(visited_sequences)} mutated sequences"
        )
        sorted_results = sorted(
            results, key=lambda result: result["score"], reverse=True
        )[:number_of_sequences]
        if len(sorted_results) < 2:
            logger.error(
                "optimization failed, could not find a mutated sequence more optimal than the original"
            )
        else:
            logger.info(
                f"found {len(sorted_results) -  1} optimal mutated sequences, best score: {sorted_results[0]['score']}"
            )
        return sorted_results



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Language modeling training utilities."""

import logging
from dataclasses import dataclass, field
from typing import Dict, Optional, Tuple, Union

import sentencepiece as _sentencepiece
from pytorch_lightning import LightningDataModule, LightningModule

from ...core import TrainingPipelineArguments
from ..core import PyTorchLightningTrainingPipeline
from .lm_datasets import CGMDataModule, CLMDataModule, MLMDataModule, PLMDataModule
from .models import LM_MODULE_FACTORY, CGMModule, CLMModule, MLMModule, PLMModule

# sentencepiece has to be loaded before lightning to avoid segfaults
_sentencepiece

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class LanguageModelingTrainingPipeline(PyTorchLightningTrainingPipeline):
    """Language modeling training pipelines."""

    def get_data_and_model_modules(
        self,
        model_args: Dict[str, Union[float, str, int]],
        dataset_args: Dict[str, Union[float, str, int]],
        **kwargs,
    ) -> Tuple[LightningDataModule, LightningModule]:
        """Get data and model modules for training.

        Args:
            model_args: model arguments passed to the configuration.
            dataset_args: dataset arguments passed to the configuration.

        Returns:
            the data and model modules.
        """

        if (
            model_args["model_config_name"] is None
            and model_args["model_name_or_path"] is None
        ):
            raise ValueError("Model config or model name/path should be provided")

        if (
            model_args["model_config_name"] is not None
            and model_args["model_name_or_path"] is not None
        ):
            logger.warning(
                "Config name is omitted. Start fine-tuning using {}".format(
                    model_args["model_name_or_path"]
                )
            )

        if model_args["tokenizer"] is None:

            if model_args["model_name_or_path"] is not None:
                model_args["tokenizer"] = model_args["model_name_or_path"]
            else:
                model_args["tokenizer"] = model_args["model_config_name"]
                logger.warning(
                    "{} tokenizer is going to be used in the training".format(
                        model_args["tokenizer"]
                    )
                )

        logger.info(f"Model arguments: {model_args}")
        logger.info(f"Dataset arguments: {dataset_args}")

        if model_args["type"] == "mlm":
            data_module, model_module = self.get_mlm_modules(model_args, dataset_args)
        elif model_args["type"] == "clm":
            data_module, model_module = self.get_clm_modules(model_args, dataset_args)  # type: ignore
        elif model_args["type"] == "plm":
            data_module, model_module = self.get_plm_modules(model_args, dataset_args)  # type: ignore
        elif model_args["type"] == "cgm":
            data_module, model_module = self.get_cgm_modules(model_args, dataset_args)  # type: ignore
        else:
            raise ValueError(f"LM training type {model_args['type']} not supported")

        model_module.model.resize_token_embeddings(len(data_module.tokenizer))  # type: ignore

        return data_module, model_module

    def get_mlm_modules(
        self,
        model_args: Dict[str, Union[float, str, int]],
        dataset_args: Dict[str, Union[float, str, int]],
    ) -> Tuple[MLMDataModule, MLMModule]:
        """Get model and data module for clm.

        Args:
            model_args: dictionary containing all the parameters for the mode configuration.
            dataset_args: dictionary containing all the necessary parameters for the dataset creation.
        Returns:
            model and data module for clm.
        """

        model_module = MLMModule(model_args)
        data_module = MLMDataModule(dataset_args, tokenizer=model_module.tokenizer)

        return data_module, model_module

    def get_clm_modules(
        self,
        model_args: Dict[str, Union[float, str, int]],
        dataset_args: Dict[str, Union[float, str, int]],
    ) -> Tuple[CLMDataModule, CLMModule]:
        """Get model and data module for clm.

        Args:
            model_args: dictionary containing all the parameters for the mode configuration.
            dataset_args: dictionary containing all the necessary parameters for the dataset creation.
        Returns:
            model and data module for clm.
        """

        model_module = CLMModule(model_args)
        data_module = CLMDataModule(dataset_args, tokenizer=model_module.tokenizer)

        return data_module, model_module

    def get_plm_modules(
        self,
        model_args: Dict[str, Union[float, str, int]],
        dataset_args: Dict[str, Union[float, str, int]],
    ) -> Tuple[PLMDataModule, PLMModule]:
        """Get model and data module for plm.

        Args:
            model_args: dictionary containing all the parameters for the mode configuration.
            dataset_args: dictionary containing all the necessary parameters for the dataset creation.
        Returns:
            model and data module for plm.
        """

        model_module = PLMModule(model_args)
        data_module = PLMDataModule(dataset_args, tokenizer=model_module.tokenizer)

        return data_module, model_module

    def get_cgm_modules(
        self,
        model_args: Dict[str, Union[float, str, int]],
        dataset_args: Dict[str, Union[float, str, int]],
    ) -> Tuple[CGMDataModule, CGMModule]:
        """Get model and data module for Conditional Generation model.

        Args:
            model_args: dictionary containing all the parameters for the mode configuration.
            dataset_args: dictionary containing all the necessary parameters for the dataset creation.
        Returns:
            model and data module for plm.
        """

        model_module = CGMModule(model_args)
        data_module = CGMDataModule(dataset_args, tokenizer=model_module.tokenizer)

        return data_module, model_module


@dataclass
class LanguageModelingModelArguments(TrainingPipelineArguments):
    """
    Arguments pertaining to which model/config we are going to fine-tune, or train from scratch.
    """

    __name__ = "model_args"

    type: str = field(
        metadata={"help": "The language modeling type, for example mlm."},
    )
    model_name_or_path: Optional[str] = field(
        default=None,
        metadata={
            "help": "The model checkpoint for weights initialization, for example bert-base-uncased."
        },
    )
    model_config_name: Optional[str] = field(
        default=None,
        metadata={"help": "Pretrained config name or path."},
    )
    tokenizer: Optional[str] = field(
        default=None,
        metadata={
            "help": "Name of the tokenizer to be used, default: tokenizer of utilizing model."
        },
    )
    lr: float = field(
        default=2e-5,
        metadata={"help": "The learning rate."},
    )
    lr_decay: float = field(
        default=0.5,
        metadata={"help": "The learning rate decay."},
    )
    cache_dir: Union[str, None] = field(
        default=None,
        metadata={
            "help": "Where do you want to store the pretrained models downloaded from huggingface.co."
        },
    )


@dataclass
class LanguageModelingDataArguments(TrainingPipelineArguments):
    """
    Arguments pertaining to what data we are going to input our model for training and eval.
    """

    __name__ = "dataset_args"

    train_file: str = field(
        metadata={
            "help": "The input training data file (a text file), for example path/to/file."
        }
    )
    validation_file: str = field(
        metadata={
            "help": "The input evaluation data file to evaluate the perplexity on (a text file), for example path/to/file."
        },
    )
    max_length: int = field(
        default=512,
        metadata={
            "help": "The maximum total input sequence length after tokenization. Sequences longer "
            "than this will be truncated."
        },
    )
    mlm_probability: float = field(
        default=0.15,
        metadata={"help": "Ratio of tokens to mask for masked language modeling loss"},
    )
    plm_probability: float = field(
        default=0.16666,
        metadata={
            "help": "Ratio of length of a span of masked tokens to surrounding context length for "
            "permutation language modeling."
        },
    )
    max_span_length: int = field(
        default=5,
        metadata={
            "help": "Maximum length of a span of masked tokens for permutation language modeling."
        },
    )
    batch_size: int = field(
        default=8,
        metadata={"help": "Ratio of tokens to mask for masked language modeling loss."},
    )


@dataclass
class LanguageModelingSavingArguments(TrainingPipelineArguments):
    """Saving arguments related to LM trainer."""

    __name__ = "saving_args"

    hf_model_path: str = field(
        metadata={"help": "Path to the converted HF model."},
        default="/tmp/gt4sd_lm_saving_tmp",
    )
    training_type: Optional[str] = field(
        metadata={
            "help": f"Training type of the converted model, supported types: {', '.join(LM_MODULE_FACTORY.keys())}."
        },
        default=None,
    )
    model_name_or_path: Optional[str] = field(
        metadata={
            "help": "Model name or path.",
        },
        default=None,
    )
    ckpt: Optional[str] = field(metadata={"help": "Path to checkpoint."}, default=None)
    tokenizer_name_or_path: Optional[str] = field(
        default=None,
        metadata={
            "help": "Tokenizer name or path. If not provided defaults to model_name_or_path."
        },
    )



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Module initialization for gt4sd."""
from ..extras import EXTRAS_ENABLED

# NOTE: here we import the applications to register them
from .conditional_generation.guacamol.core import (  # noqa: F401
    AaeGenerator,
    GraphGAGenerator,
    GraphMCTSGenerator,
    OrganGenerator,
    SMILESGAGenerator,
    SMILESLSTMHCGenerator,
    SMILESLSTMPPOGenerator,
    VaeGenerator,
)
from .conditional_generation.key_bert.core import KeyBERTGenerator  # noqa: F401
from .conditional_generation.paccmann_rl.core import (  # noqa: F401
    PaccMannRLOmicBasedGenerator,
    PaccMannRLProteinBasedGenerator,
)
from .conditional_generation.regression_transformer.core import (  # noqa: F401
    RegressionTransformerMolecules,
    RegressionTransformerProteins,
)
from .conditional_generation.reinvent.core import ReinventGenerator  # noqa: F401
from .conditional_generation.template.core import TemplateGenerator  # noqa: F401
from .controlled_sampling.advanced_manufacturing.core import (  # noqa: F401
    CatalystGenerator,
)
from .controlled_sampling.paccmann_gp.core import PaccMannGPGenerator  # noqa: F401
from .generation.diffusion.core import (  # noqa: F401
    DDIMGenerator,
    DDPMGenerator,
    LDMGenerator,
    LDMTextToImageGenerator,
    ScoreSdeGenerator,
    StableDiffusionGenerator,
)
from .generation.hugging_face.core import (  # noqa: F401
    HuggingFaceCTRLGenerator,
    HuggingFaceGPT2Generator,
    HuggingFaceOpenAIGPTGenerator,
    HuggingFaceTransfoXLGenerator,
    HuggingFaceXLMGenerator,
    HuggingFaceXLNetGenerator,
)
from .generation.moler.core import MoLeRDefaultGenerator  # noqa: F401
from .generation.pgt.core import (  # noqa: F401
    PGTCoherenceChecker,
    PGTEditor,
    PGTGenerator,
)
from .generation.polymer_blocks.core import PolymerBlocksGenerator  # noqa: F401
from .generation.torchdrug.core import TorchDrugGCPN, TorchDrugGraphAF  # noqa: F401
from .prediction.paccmann.core import PaccMann  # noqa: F401
from .prediction.topics_zero_shot.core import TopicsPredictor  # noqa: F401

# extras requirements
if EXTRAS_ENABLED:
    from .controlled_sampling.class_controlled_sampling.core import (  # noqa: F401
        PAG,
        CogMol,
    )
    from .generation.molgx.core import MolGXQM9Generator  # noqa: F401



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Template Algorithm"""

import logging
from dataclasses import field
from typing import Any, Callable, ClassVar, Dict, Iterable, Optional, TypeVar

from ...core import AlgorithmConfiguration, GeneratorAlgorithm  # type: ignore
from ...registry import ApplicationsRegistry  # type: ignore
from .implementation import Generator  # type: ignore

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

T = TypeVar("T")
S = TypeVar("S")
Targeted = Callable[[T], Iterable[Any]]


class Template(GeneratorAlgorithm[S, T]):
    """Template Algorithm."""

    def __init__(
        self, configuration: AlgorithmConfiguration[S, T], target: Optional[T] = None
    ):
        """Template Generation

        Args:
            configuration: domain and application
                specification, defining types and validations.
            target: Optional depending on the type of generative model. In this template
                we will convert the target to a string.

        Example:
            An example for using this template::

            target = 'World'
            configuration = TemplateGenerator()
            algorithm = Template(configuration=configuration, target=target)
            items = list(algorithm.sample(1))
            print(items)
        """

        configuration = self.validate_configuration(configuration)
        # TODO there might also be a validation/check on the target input

        super().__init__(
            configuration=configuration,
            target=target,  # type:ignore
        )

    def get_generator(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ) -> Targeted[T]:
        """Get the function to from generator.

        Args:
            configuration: helps to set up the application.
            target: context or condition for the generation. Just an optional string here.

        Returns:
            callable generating a list of 1 item containing salutation and temperature converted to fahrenheit.
        """
        logger.info("ensure artifacts for the application are present.")
        self.local_artifacts = configuration.ensure_artifacts()
        implementation: Generator = configuration.get_conditional_generator(  # type: ignore
            self.local_artifacts
        )
        return implementation.hello_name  # type:ignore

    def validate_configuration(
        self, configuration: AlgorithmConfiguration
    ) -> AlgorithmConfiguration:
        # TODO raise InvalidAlgorithmConfiguration
        assert isinstance(configuration, AlgorithmConfiguration)
        return configuration


@ApplicationsRegistry.register_algorithm_application(Template)
class TemplateGenerator(AlgorithmConfiguration[str, str]):
    """Configuration for specific generator."""

    algorithm_type: ClassVar[str] = "conditional_generation"
    domain: ClassVar[str] = "materials"
    algorithm_version: str = "v0"

    temperature: int = field(
        default=36,
        metadata=dict(description="Temperature parameter ( in celsius )"),
    )

    def get_target_description(self) -> Dict[str, str]:
        """Get description of the target for generation.
        Returns:
            target description.
        """
        return {
            "title": "Target name",
            "description": "A simple string to define the name in the output [Hello name].",
            "type": "string",
        }

    def get_conditional_generator(self, resources_path: str) -> Generator:
        return Generator(resources_path=resources_path, temperature=self.temperature)



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""PaccMann\\ :superscript:`RL` Algorithm.

PaccMann\\ :superscript:`RL` generation is conditioned via reinforcement learning.
"""

import logging
from dataclasses import field
from typing import Any, Callable, ClassVar, Dict, Iterable, Optional, TypeVar

from typing_extensions import Protocol, runtime_checkable

from ....domains.materials import SMILES, Omics, Protein
from ....exceptions import InvalidItem
from ...core import AlgorithmConfiguration, GeneratorAlgorithm
from ...registry import ApplicationsRegistry
from .implementation import (
    ConditionalGenerator,
    ProteinSequenceConditionalGenerator,
    TranscriptomicConditionalGenerator,
)

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

T = TypeVar("T", Protein, Omics)
S = TypeVar("S", bound=SMILES)
Targeted = Callable[[T], Iterable[Any]]


class PaccMannRL(GeneratorAlgorithm[S, T]):
    """PaccMann\\ :superscript:`RL` Algorithm."""

    def __init__(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ):
        """Instantiate PaccMannRL ready to generate items.

        Args:
            configuration: domain and application
                specification defining parameters, types and validations.
            target: a target for which to generate items.

        Example:
            An example for generating small molecules (SMILES) with high affinity
            for a target protein::

                affinity_config = PaccMannRLProteinBasedGenerator()
                target = "MVLSPADKTNVKAAWGKVGAHAGEYGAEALERMFLSFPTT"
                paccmann_affinity = PaccMannRL(configuration=affinity_config, target=target)
                items = list(paccmann_affinity.sample(10))
                print(items)
        """

        configuration = self.validate_configuration(configuration)
        # TODO there might also be a validation/check on the target input

        super().__init__(
            configuration=configuration,  # type:ignore
            target=target,  # type:ignore
        )

    def get_generator(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ) -> Targeted[T]:
        """Get the function to sample batches via PaccMannRL's ConditionalGenerator.

        Args:
            configuration: helps to set up specific application of PaccMannRL.
            target: context or condition for the generation.

        Returns:
            callable with target generating a batch of items.
        """
        logger.info("ensure artifacts for the application are present.")
        self.local_artifacts = configuration.ensure_artifacts()
        implementation: ConditionalGenerator = configuration.get_conditional_generator(  # type: ignore
            self.local_artifacts
        )
        return implementation.generate_batch

    def validate_configuration(
        self, configuration: AlgorithmConfiguration[S, T]
    ) -> AlgorithmConfiguration[S, T]:
        @runtime_checkable
        class AnyPaccMannRLConfiguration(Protocol):
            """Protocol for PaccMannRL configurations."""

            def get_conditional_generator(
                self, resources_path: str
            ) -> ConditionalGenerator:
                ...

            def validate_item(self, item: Any) -> S:
                ...

        # TODO raise InvalidAlgorithmConfiguration
        assert isinstance(configuration, AnyPaccMannRLConfiguration)
        assert isinstance(configuration, AlgorithmConfiguration)
        return configuration


@ApplicationsRegistry.register_algorithm_application(PaccMannRL)
class PaccMannRLProteinBasedGenerator(AlgorithmConfiguration[SMILES, Protein]):
    """
    Configuration to generate compounds with high affinity to a target protein.

    Implementation from the paper: https://doi.org/10.1088/2632-2153/abe808.
    """

    algorithm_type: ClassVar[str] = "conditional_generation"
    domain: ClassVar[str] = "materials"
    algorithm_version: str = "v0"

    batch_size: int = field(
        default=32,
        metadata=dict(description="Batch size used for the generative model sampling."),
    )
    temperature: float = field(
        default=1.4,
        metadata=dict(
            description="Temperature parameter for the softmax sampling in decoding."
        ),
    )
    generated_length: int = field(
        default=100,
        metadata=dict(
            description="Maximum length in tokens of the generated molcules (relates to the SMILES length)."
        ),
    )

    def get_target_description(self) -> Dict[str, str]:
        """Get description of the target for generation.

        Returns:
            target description.
        """
        return {
            "title": "Target protein sequence",
            "description": "AA sequence of the protein target to generate non-toxic ligands against.",
            "type": "string",
        }

    def get_conditional_generator(
        self, resources_path: str
    ) -> ProteinSequenceConditionalGenerator:
        """Instantiate the actual generator implementation.

        Args:
            resources_path: local path to model files.

        Returns:
            instance with :meth:`generate_batch<gt4sd.algorithms.conditional_generation.paccmann_rl.implementation.ConditionalGenerator.generate_batch>` method for targeted generation.
        """
        return ProteinSequenceConditionalGenerator(
            resources_path=resources_path,
            temperature=self.temperature,
            generated_length=self.generated_length,
            samples_per_protein=self.batch_size,
        )

    def validate_item(self, item: str) -> SMILES:
        """Check that item is a valid SMILES.

        Args:
            item: a generated item that is possibly not valid.

        Raises:
            InvalidItem: in case the item can not be validated.

        Returns:
            the validated SMILES.
        """
        (
            molecules,
            _,
        ) = ProteinSequenceConditionalGenerator.validate_molecules([item])
        if molecules[0] is None:
            raise InvalidItem(
                title="InvalidSMILES",
                detail=f'rdkit.Chem.MolFromSmiles returned None for "{item}"',
            )
        return SMILES(item)


@ApplicationsRegistry.register_algorithm_application(PaccMannRL)
class PaccMannRLOmicBasedGenerator(AlgorithmConfiguration[SMILES, Omics]):
    """
    Configuration to generate compounds with low IC50 for a target omics profile.

    Implementation from the paper: https://doi.org/10.1016/j.isci.2021.102269.
    """

    algorithm_type: ClassVar[str] = "conditional_generation"
    domain: ClassVar[str] = "materials"
    algorithm_version: str = "v0"

    batch_size: int = field(
        default=32,
        metadata=dict(description="Batch size used for the generative model sampling."),
    )
    temperature: float = field(
        default=1.4,
        metadata=dict(
            description="Temperature parameter for the softmax sampling in decoding."
        ),
    )
    generated_length: int = field(
        default=100,
        metadata=dict(
            description="Maximum length in tokens of the generated molcules (relates to the SMILES length)."
        ),
    )

    def get_target_description(self) -> Dict[str, str]:
        """Get description of the target for generation.

        Returns:
            target description.
        """
        return {
            "title": "Gene expression profile",
            "description": "A gene expression profile to generate effective molecules against.",
            "type": "list",
        }

    def get_conditional_generator(
        self, resources_path: str
    ) -> TranscriptomicConditionalGenerator:
        """Instantiate the actual generator implementation.

        Args:
            resources_path: local path to model files.

        Returns:
            instance with :meth:`generate_batch<gt4sd.algorithms.conditional_generation.paccmann_rl.implementation.ConditionalGenerator.generate_batch>` method for targeted generation.
        """
        return TranscriptomicConditionalGenerator(
            resources_path=resources_path,
            temperature=self.temperature,
            generated_length=self.generated_length,
            samples_per_profile=self.batch_size,
        )

    def validate_item(self, item: str) -> SMILES:
        """Check that item is a valid SMILES.

        Args:
            item: a generated item that is possibly not valid.

        Raises:
            InvalidItem: in case the item can not be validated.

        Returns:
            the validated SMILES.
        """
        (
            molecules,
            _,
        ) = TranscriptomicConditionalGenerator.validate_molecules([item])
        if molecules[0] is None:
            raise InvalidItem(
                title="InvalidSMILES",
                detail=f'rdkit.Chem.MolFromSmiles returned None for "{item}"',
            )
        return SMILES(item)



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Argument parser for training pipelines."""


import ast
import dataclasses
import re
from argparse import ArgumentTypeError
from enum import Enum
from functools import partial
from typing import Any, Callable, List, NewType, Optional, Type, Union

from transformers import HfArgumentParser


def eval_lambda(val: str) -> Callable:
    """Parse a lambda from a string safely.

    Args:
        val: string representing a lambda.

    Returns:
        a callable.

    Raises:
        ValueError: in case the lambda can not be parsed.
    """
    parsed_lamba = ast.parse(val).body[0].value  # type:ignore
    if isinstance(parsed_lamba, ast.Lambda) and "eval" not in val:
        return eval(val)
    else:
        raise ValueError(f"'{val}' can not be safely parsed as a lambda function")


def none_checker_bool(val: Union[bool, str]) -> Union[bool, None]:
    """Check given bool argument for None.

    Args:
        val: model arguments passed to the configuration.

    Returns:
        Bool value or None.

    Raises:
        ArgumentTypeError: value can not be parsed.
    """
    if not val:
        return None
    if isinstance(val, bool):
        return val
    if val.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif val.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise ArgumentTypeError(
            f"Truthy value expected: got {val} but expected one of yes/no, true/false, t/f, y/n, 1/0 (case insensitive)."
        )


def none_checker(val: Any, dtype: Type) -> Any:
    """Check given argument for None.

    Args:
        val: model arguments passed to the configuration.
        dtype: expected argument type.

    Returns:
        Value casted in the expected type or None.
    """
    if not val or val == "none":
        return None
    return dtype(val)


DataClass = NewType("DataClass", Any)  # type: ignore
DataClassType = NewType("DataClassType", Any)  # type: ignore


class ArgumentParser(HfArgumentParser):
    """ArgumentParser inherited from hf's parser with modified dataclass arguments addition for better handling of None values."""

    def _add_dataclass_arguments(self, dtype: DataClassType) -> None:
        """Add a dataclass arguments.

        Args:
            dtype: data class type.
        """

        if hasattr(dtype, "_argument_group_name"):
            parser = self.add_argument_group(dtype._argument_group_name)
        else:
            parser = self  # type: ignore
        for field in dataclasses.fields(dtype):
            if not field.init:
                continue
            field_name = f"--{field.name}"
            kwargs = field.metadata.copy()  # type: ignore
            # field.metadata is not used at all by Data Classes,
            # it is provided as a third-party extension mechanism.
            if isinstance(field.type, str):
                raise ImportError(
                    "This implementation is not compatible with Postponed Evaluation of Annotations (PEP 563),"
                    "which can be opted in from Python 3.7 with `from __future__ import annotations`."
                    "We will add compatibility when Python 3.9 is released."
                )
            typestring = str(field.type)
            for prim_type in (int, float, str):
                for collection in (List,):
                    if (
                        typestring == f"typing.Union[{collection[prim_type]}, NoneType]"  # type: ignore
                        or typestring == f"typing.Optional[{collection[prim_type]}]"  # type: ignore
                    ):
                        field.type = collection[prim_type]  # type: ignore
                if (
                    typestring == f"typing.Union[{prim_type.__name__}, NoneType]"
                    or typestring == f"typing.Optional[{prim_type.__name__}]"
                ):
                    field.type = prim_type

            if isinstance(field.type, type) and issubclass(field.type, Enum):
                kwargs["choices"] = [x.value for x in field.type]
                kwargs["type"] = type(kwargs["choices"][0])
                if field.default is not dataclasses.MISSING:
                    kwargs["default"] = field.default
                else:
                    kwargs["required"] = True
            elif field.type is bool or field.type == Optional[bool]:

                if field.default is True:
                    parser.add_argument(
                        f"--no_{field.name}",
                        action="store_false",
                        dest=field.name,
                        **kwargs,
                    )

                # Hack because type=bool in argparse does not behave as we want.
                kwargs["type"] = none_checker_bool
                if field.type is bool or (
                    field.default is not None
                    and field.default is not dataclasses.MISSING
                ):
                    # Default value is False if we have no default when of type bool.
                    default = (
                        False if field.default is dataclasses.MISSING else field.default
                    )
                    # This is the value that will get picked if we don't include --field_name in any way
                    kwargs["default"] = default
                    # This tells argparse we accept 0 or 1 value after --field_name
                    kwargs["nargs"] = "?"
                    # This is the value that will get picked if we do --field_name (without value)
                    kwargs["const"] = True
            elif (
                hasattr(field.type, "__origin__")
                and re.search(r"^typing\.List\[(.*)\]$", str(field.type)) is not None
            ):
                kwargs["nargs"] = "+"
                kwargs["type"] = partial(none_checker, dtype=field.type.__args__[0])
                assert all(
                    x == kwargs["type"] for x in field.type.__args__
                ), f"{field.name} cannot be a List of mixed types: {field.type.__args__}"
                if field.default_factory is not dataclasses.MISSING:  # type: ignore
                    kwargs["default"] = field.default_factory()  # type: ignore
                elif field.default is dataclasses.MISSING:
                    kwargs["required"] = True
            else:
                kwargs["type"] = partial(none_checker, dtype=field.type)
                if field.default is not dataclasses.MISSING:
                    kwargs["default"] = field.default
                elif field.default_factory is not dataclasses.MISSING:  # type: ignore
                    kwargs["default"] = field.default_factory()  # type: ignore
                else:
                    kwargs["required"] = True
            parser.add_argument(field_name, **kwargs)



#!/usr/bin/env python
#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Run model saving for the GT4SD."""

import logging
import os
import sys
from dataclasses import dataclass, field
from typing import IO, Iterable, Optional, cast

from ..algorithms.registry import ApplicationsRegistry
from ..training_pipelines import TRAINING_PIPELINE_ARGUMENTS_FOR_MODEL_SAVING
from ..training_pipelines.core import TrainingPipelineArguments
from .algorithms import (
    AVAILABLE_ALGORITHMS,
    AVAILABLE_ALGORITHMS_CATEGORIES,
    filter_algorithm_applications,
    get_configuration_tuples,
)
from .argument_parser import ArgumentParser, DataClassType

logger = logging.getLogger(__name__)

SUPPORTED_TRAINING_PIPELINES = sorted(
    TRAINING_PIPELINE_ARGUMENTS_FOR_MODEL_SAVING.keys()
)


@dataclass
class SavingArguments:
    """Algorithm saving arguments."""

    __name__ = "saving_base_args"

    training_pipeline_name: str = field(
        metadata={
            "help": f"Training pipeline name, supported pipelines: {', '.join(SUPPORTED_TRAINING_PIPELINES)}."
        },
    )
    target_version: str = field(
        metadata={"help": "Target algorithm version to save."},
    )
    algorithm_type: Optional[str] = field(
        default=None,
        metadata={
            "help": f"Inference algorithm type, supported types: {', '.join(AVAILABLE_ALGORITHMS_CATEGORIES['algorithm_type'])}."
        },
    )
    domain: Optional[str] = field(
        default=None,
        metadata={
            "help": f"Domain of the inference algorithm, supported types: {', '.join(AVAILABLE_ALGORITHMS_CATEGORIES['domain'])}."
        },
    )
    algorithm_name: Optional[str] = field(
        default=None,
        metadata={"help": "Inference algorithm name."},
    )
    algorithm_application: Optional[str] = field(
        default=None,
        metadata={"help": "Inference algorithm application."},
    )
    source_version: Optional[str] = field(
        default=None,
        metadata={"help": "Source algorithm version to use for missing artifacts."},
    )


class SavingArgumentParser(ArgumentParser):
    """Argument parser using a custom help logic."""

    def print_help(self, file: Optional[IO[str]] = None) -> None:
        """Print help checking dynamically whether a specific pipeline is passed.

        Args:
            file: an optional I/O stream. Defaults to None, a.k.a., stdout and stderr.
        """
        try:
            help_args_set = {"-h", "--help"}
            if (
                len(set(sys.argv).union(help_args_set)) < len(help_args_set) + 2
            ):  # considering filename
                super().print_help()
                return
            args = [arg for arg in sys.argv if arg not in help_args_set]
            parsed_arguments = super().parse_args_into_dataclasses(
                args=args, return_remaining_strings=True
            )
            trainer_arguments = None
            for arguments in parsed_arguments:
                if arguments.__name__ == "trainer_base_args":
                    trainer_arguments = arguments
                    break
            if trainer_arguments:
                trainer_arguments.training_pipeline_name
                training_pipeline_arguments = (
                    TRAINING_PIPELINE_ARGUMENTS_FOR_MODEL_SAVING.get(
                        trainer_arguments.training_pipeline_name,
                        TrainingPipelineArguments,
                    )
                )
                parser = ArgumentParser(
                    tuple(
                        [SavingArguments, *training_pipeline_arguments]  # type:ignore
                    )
                )
                parser.print_help()
        except Exception:
            super().print_help()


def main() -> None:
    """
    Run an algorithm saving pipeline.

    Raises:
        ValueError: in case the provided training pipeline provided is not supported.
    """
    logging.basicConfig(stream=sys.stdout, level=logging.INFO)

    base_args = SavingArgumentParser(
        cast(DataClassType, SavingArguments)
    ).parse_args_into_dataclasses(return_remaining_strings=True)[0]
    training_pipeline_name = base_args.training_pipeline_name
    if training_pipeline_name not in set(SUPPORTED_TRAINING_PIPELINES):
        ValueError(
            f"Training pipeline {training_pipeline_name} is not supported. Supported types: {', '.join(SUPPORTED_TRAINING_PIPELINES)}."
        )
    training_pipeline_saving_arguments = TRAINING_PIPELINE_ARGUMENTS_FOR_MODEL_SAVING[
        training_pipeline_name
    ]
    parser = SavingArgumentParser(
        cast(
            Iterable[DataClassType],
            tuple([SavingArguments, training_pipeline_saving_arguments]),
        )
    )

    saving_args, training_pipeline_saving_args, _ = parser.parse_args_into_dataclasses(
        return_remaining_strings=True
    )

    filters = {
        key: saving_args.__dict__[key]
        for key in [
            "algorithm_type",
            "algorithm_application",
            "domain",
            "algorithm_name",
            "source_version",
        ]
    }
    configuration_tuples = get_configuration_tuples(
        filter_algorithm_applications(algorithms=AVAILABLE_ALGORITHMS, filters=filters)
    )
    if len(configuration_tuples) > 1:
        logger.info(
            f"Multiple configurations matching the parameters:{os.linesep}"
            f"{os.linesep.join(map(str, configuration_tuples))}{os.linesep}"
            f"Select one by specifying additional algorithms parameters: {','.join('--' + key for key, value in filters.items() if not value)}.",
        )
        return
    elif len(configuration_tuples) < 1:
        provided_filters = {key: value for key, value in filters.items() if value}
        logger.error(
            "No configurations matching the provided parameters, "
            f"please review the supported configurations:{os.linesep}"
            f"{os.linesep.join(map(str, configuration_tuples))}{os.linesep}"
            f"Please review the parameters provided:{os.linesep}"
            f"{provided_filters}"
        )
    configuration_tuple = configuration_tuples[0]
    logger.info(f"Selected configuration: {configuration_tuple}")

    algorithm_application = ApplicationsRegistry.applications[configuration_tuple]
    configuration_class = algorithm_application.configuration_class
    logger.info(
        f'Saving model version "{saving_args.target_version}" with the following configuration: {configuration_class}'
    )
    configuration_class.save_version_from_training_pipeline_arguments(
        training_pipeline_arguments=training_pipeline_saving_args,
        target_version=saving_args.target_version,
        source_version=saving_args.source_version,
    )


if __name__ == "__main__":
    main()



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Dataset routines-filtering, dataset building."""

import json
import logging
import os
from functools import lru_cache
from typing import Any, Callable, Dict, List, Union

import sentencepiece as _sentencepiece
import pytorch_lightning as pl
from datasets import DatasetDict
from torch.utils.data import ConcatDataset, DataLoader, Dataset
from transformers import (
    AutoTokenizer,
    DataCollatorForLanguageModeling,
    DataCollatorForPermutationLanguageModeling,
    default_data_collator,
)
from transformers.tokenization_utils_base import BatchEncoding

# Sentencepiece has to be loaded before lightning
_sentencepiece

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class LMDataset(Dataset):
    """LM dataset class."""

    def __init__(
        self,
        filepath: str,
        tokenizer: Callable,
    ) -> None:
        """Initialize the LM data module.

        Args:
            filepath: path where the dataset is located.
            tokenizer: tokenize function to be used in the module.
        """

        self.filepath = filepath
        self.tokenizer = tokenizer
        self.length = LMDataset.count_examples(filepath)

        if not self.filepath.endswith(".jsonl") and not self.filepath.endswith(".json"):
            raise ValueError(f"{filepath} is not a .jsonl or a json.")

    @lru_cache()
    def examples_reader(self) -> List[Dict[str, str]]:
        """Read instances from a filepath.

        Returns:
           list of instances.
        """
        with open(self.filepath) as fp:
            return [json.loads(line.strip()) for line in fp]

    @staticmethod
    def count_examples(filepath: str) -> int:
        """Count instances of a filepath.

        Args:
           filepath: path of the dataset.
        Returns:
           number of examples existed in the given filepath.
        """

        def _make_gen(reader):
            while True:
                b = reader(2**16)
                if not b:
                    break
                yield b

        with open(filepath, "rb") as f:
            count = sum(buf.count(b"\n") for buf in _make_gen(f.raw.read))  # type: ignore
        return count

    def __len__(self) -> int:
        """Number of instances of the dataset.

        Returns:
           number of instances
        """
        return self.length

    def __getitem__(self, index) -> BatchEncoding:
        """Get an item of the dataset.

        Args:
            index: index of the item.
        Returns:
            tokenized item.
        """

        examples = self.examples_reader()
        example = self.tokenizer(examples[index])

        return example


class DataModule(pl.LightningDataModule):
    """Pytorch-lightning-style data module for LM dataset."""

    def __init__(self, dataset_args: Dict[str, Any], tokenizer: AutoTokenizer) -> None:
        """Initialize the data module.

        Args:
            dataset_args: dictionary containing the arguments for the lightning data module creation.
            tokenizer: tokenizer to be used in the module.
        """

        super().__init__()

        self.dataset: DatasetDict

        self.dataset_args = dataset_args

        self.tokenizer = tokenizer

        self.data_collator = default_data_collator

        if "num_dataloader_workers" not in self.dataset_args:

            self.dataset_args["num_dataloader_workers"] = 8

            cpus_count = os.cpu_count()
            if cpus_count is not None:
                self.dataset_args["num_dataloader_workers"] = min(
                    self.dataset_args["num_dataloader_workers"], cpus_count
                )

    def build_dataset(self, path: str) -> Dataset:
        """
        Build the dataset.

        Args:
            path: path where the dataset is located.
        Returns:
            a torch Dataset.
        """

        if path.endswith(".jsonl") or path.endswith(".json"):
            return LMDataset(path, self.tokenize_function)
        elif os.path.isdir(path):
            return ConcatDataset(
                datasets=[
                    LMDataset(os.path.join(path, filename), self.tokenize_function)
                    for filename in os.listdir(path)
                    if filename.endswith(".jsonl") or filename.endswith(".json")
                ]
            )
        else:
            raise TypeError(f"{path} type is not supported for dataset")

    def tokenize_function(
        self, examples: Dict[str, Union[int, slice]]
    ) -> BatchEncoding:
        """Tokenize the given examples.

        Args:
            examples: list of examples.
        Returns:
            tokenized examples.
        """

        truncation = self.dataset_args.get("truncation", True)
        padding = self.dataset_args.get("padding", "max_length")
        max_length = self.dataset_args.get("max_length", 512)

        return self.tokenizer(  # type: ignore
            examples["text"],
            truncation=truncation,
            padding=padding,
            max_length=max_length,
        )

    def load(self) -> None:
        """Load datasets from the given files."""

        self.datasets = {
            "train": self.build_dataset(self.dataset_args["train_file"]),
            "validation": self.build_dataset(self.dataset_args["validation_file"]),
        }

        logger.info(
            f"Training set size: {len(self.datasets['train'])} - Validation set size: {len(self.datasets['validation'])}"  # type: ignore
        )

    def train_dataloader(self) -> DataLoader:
        """Create the DataLoader for the traning step.

        Returns:
            pytorch-like dataloader.
        """
        return DataLoader(
            self.datasets["train"],  # type: ignore
            batch_size=self.dataset_args["batch_size"],
            num_workers=self.dataset_args["num_dataloader_workers"],
            collate_fn=self.data_collator,
        )

    def val_dataloader(self) -> DataLoader:
        """Create the DataLoader for the traning step.

        Returns:
            pytorch-like dataloader.
        """
        return DataLoader(
            self.datasets["validation"],  # type: ignore
            batch_size=self.dataset_args["batch_size"],
            num_workers=self.dataset_args["num_dataloader_workers"],
            collate_fn=self.data_collator,
        )


class MLMDataModule(DataModule):
    """Pytorch-lightning-style data module for MLM dataset."""

    def __init__(
        self, dataset_args: Dict[str, Union[float, str, int]], tokenizer: AutoTokenizer
    ) -> None:
        """Initialize the data module.

        Args:
            dataset_args: dictionary containing the metadata for the lightning data module creation.
            tokenizer: tokenizer to be used in the module.
        """
        super().__init__(dataset_args, tokenizer)

        self.data_collator = DataCollatorForLanguageModeling(
            self.tokenizer, self.dataset_args["mlm_probability"]  # type: ignore
        )

        self.load()


class CGMDataModule(DataModule):
    """Pytorch-lightning-style data module for conditional generation dataset."""

    def __init__(
        self, dataset_args: Dict[str, Union[float, str, int]], tokenizer: AutoTokenizer
    ) -> None:
        """
        Initialize the data module.

        Args:
            dataset_args: dictionary containing the metadata for the lightning data module creation.
            tokenizer: tokenizer to be used in the module.
        """
        super().__init__(dataset_args, tokenizer)

        self.load()

    def tokenize_function(
        self, examples: Dict[str, Union[int, slice]]
    ) -> BatchEncoding:
        """Tokenize the given examples.

        Args:
            examples: list of examples.
        Returns:
            tokenized examples.
        """

        truncation = self.dataset_args.get("truncation", True)
        padding = self.dataset_args.get("padding", "max_length")
        max_length = self.dataset_args.get("max_length", 512)

        source = self.tokenizer(  # type: ignore
            examples["source"],
            truncation=truncation,
            padding=padding,
            max_length=max_length,
        )

        targets = self.tokenizer(  # type: ignore
            examples["target"],
            truncation=truncation,
            padding=padding,
            max_length=max_length,
        )

        return BatchEncoding(
            data={
                "input_ids": source["input_ids"],
                "attention_mask": source["attention_mask"],
                "labels": targets["input_ids"],
                "decoder_attention_mask": targets["attention_mask"],
            }
        )


class CLMDataModule(DataModule):
    """Pytorch-lightning-style data module for CLM dataset."""

    def __init__(
        self, dataset_args: Dict[str, Union[float, str, int]], tokenizer: AutoTokenizer
    ) -> None:
        """
        Initialize the data module.

        Args:
            dataset_args: dictionary containing the metadata for the lightning data module creation.
            tokenizer: tokenizer to be used in the module.
        """
        super().__init__(dataset_args, tokenizer)

        self.load()

    def tokenize_function(
        self, examples: Dict[str, Union[int, slice]]
    ) -> BatchEncoding:
        """Tokenize the given examples.

        Args:
            examples: list of examples.
        Returns:
            tokenized examples.
        """

        tokenized_data = super().tokenize_function(examples)

        tokenized_data["labels"] = tokenized_data["input_ids"].copy()

        return tokenized_data


class PLMDataModule(DataModule):
    """Pytorch-lightning-style data module for PLM dataset."""

    def __init__(
        self, dataset_args: Dict[str, Union[float, str, int]], tokenizer: AutoTokenizer
    ) -> None:
        """Initialize the data module.

        Args:
            dataset_args: dictionary containing the metadata for the lightning data module creation.
            tokenizer: tokenizer to be used in the module.
        """
        super().__init__(dataset_args, tokenizer)

        self.data_collator = DataCollatorForPermutationLanguageModeling(
            tokenizer=self.tokenizer,  # type: ignore
            plm_probability=self.dataset_args["plm_probability"],
            max_span_length=self.dataset_args["max_span_length"],
        )

        self.load()



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""GuacaMol LSTM trainer unit tests."""

import os
import shutil
import tempfile
from typing import Any, Dict, cast

import pkg_resources

from gt4sd.training_pipelines import (
    TRAINING_PIPELINE_MAPPING,
    GuacaMolLSTMTrainingPipeline,
)

TEST_DATA_DIRECTORY = pkg_resources.resource_filename(
    "gt4sd",
    "training_pipelines/tests/",
)

template_config = {
    "training_args": {
        "batch_size": 512,
        "valid_every": 1000,
        "n_epochs": 10,
        "lr": 1e-3,
    },
    "model_args": {
        "max_len": 100,
        "hidden_size": 512,
        "n_layers": 2,
        "rnn_dropout": 0.2,
    },
    "dataset_args": {},
}


def test_train():

    pipeline = TRAINING_PIPELINE_MAPPING.get("guacamol-lstm-trainer")

    assert pipeline is not None

    TEMPORARY_DIRECTORY = tempfile.mkdtemp()

    test_pipeline = cast(GuacaMolLSTMTrainingPipeline, pipeline())

    config: Dict[str, Any] = template_config.copy()
    file_path = os.path.join(TEST_DATA_DIRECTORY, "molecules.smiles")
    config["training_args"]["output_dir"] = TEMPORARY_DIRECTORY
    config["dataset_args"]["train_smiles_filepath"] = file_path
    config["dataset_args"]["test_smiles_filepath"] = file_path

    test_pipeline.train(**config)

    shutil.rmtree(TEMPORARY_DIRECTORY)
